"""
The ExploitationFeaturesType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import datetime
from typing import Union, List

import numpy

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import SerializableDescriptor, ParametersDescriptor, \
    FloatDescriptor, FloatModularDescriptor, StringDescriptor, StringEnumDescriptor, \
    DateTimeDescriptor, SerializableListDescriptor
from sarpy.io.complex.sicd_elements.SCPCOA import GeometryCalculator
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.geometry.geocoords import wgs_84_norm, ecf_to_geodetic
import sarpy.geometry.point_projection
import sarpy.processing.sicd.windows

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import RowColIntType, RowColDoubleType, RangeAzimuthType, \
    AngleMagnitudeType, RadarModeType


TX_POLARIZATION_VALUES = ('V', 'H', 'RHC', 'LHC', 'OTHER', 'UNKNOWN', 'SEQUENCE')
RCV_POLARIZATION_VALUES = ('V', 'H', 'RHC', 'LHC', 'OTHER', 'UNKNOWN')
logger = logging.getLogger(__name__)

_sicd_type_text = 'Requires instance of SICDType,\n\tgot type `{}`'
_exp_calc_text = 'Requires input which is an instance of ExploitationCalculator,\n\tgot type `{}`'


class ExploitationCalculator(object):
    """
    Helper class for calculating the various geometric values for exploitation
    features. This is predominantly using the ontology presented in the SIDD standards
    document.
    """

    def __init__(self, geometry_calculator, row_vector, col_vector):
        """

        Parameters
        ----------
        geometry_calculator : GeometryCalculator
        row_vector : numpy.ndarray
        col_vector : numpy.ndarray
        """

        # extract sicd based parameters
        self.geometry_calculator = geometry_calculator
        self.ARPPos = geometry_calculator.ARP
        self.ARPVel = geometry_calculator.ARP_vel
        self.SCP = geometry_calculator.SCP
        self.slant_x = self._make_unit(self.ARPPos - self.SCP)
        self.slant_z = self._make_unit(numpy.cross(self.slant_x, self.ARPVel))
        if self.slant_z.dot(self.ARPPos) < 0:
            self.slant_z *= -1
        self.slant_y = self._make_unit(numpy.cross(self.slant_z, self.slant_x))
        self.ETP = wgs_84_norm(self.SCP)

        # store SIDD grid parameters
        self.row_vector = self._make_unit(row_vector)
        self.col_vector = self._make_unit(col_vector)
        self.normal_vector = self._make_unit(numpy.cross(self.row_vector, self.col_vector))

    @staticmethod
    def _make_unit(vec):
        vec_norm = numpy.linalg.norm(vec)
        if vec_norm < 1e-6:
            logger.error(
                'input vector to be normalized has norm {}, this may be a mistake'.format(vec_norm))
        return vec/vec_norm

    # The geometry properties
    @property
    def AzimuthAngle(self):
        return self.geometry_calculator.AzimAng

    @property
    def SlopeAngle(self):
        """
        float: The angle between the ground plane and slant plane. Note that the
        standard outlines this is angle between the Earth Tangent Plane and slant
        plane, with the expectation that the "ground plane" will always be defined
        as the Earth Tangent Plane.
        """

        return numpy.rad2deg(numpy.arccos(self.slant_z.dot(self.ETP)))

    @property
    def DopplerConeAngle(self):
        return self.geometry_calculator.DopplerConeAng

    @property
    def SquintAngle(self):
        return self.geometry_calculator.SquintAngle

    @property
    def GrazeAngle(self):
        """
        float: The angle between the ground plane and line of sight vector.
        """

        return numpy.rad2deg(numpy.arcsin(self.slant_x.dot(self.ETP)))

    @property
    def TiltAngle(self):
        """
        float: The angle between the ground plane and cross range vector.
        """

        return numpy.rad2deg(numpy.arctan(self.ETP.dot(self.slant_y)/self.ETP.dot(self.slant_z)))

    # The phenomenology properties
    @property
    def Shadow(self):
        # type: () -> AngleMagnitudeType
        """
        AngleMagnitudeType: The shadow angle and magnitude.
        """

        shadow = self.ETP - self.slant_x/(self.slant_x.dot(self.ETP))
        shadow_prime = shadow - (shadow.dot(self.normal_vector)/self.slant_z.dot(self.normal_vector))*self.slant_z
        return AngleMagnitudeType(
            Angle=numpy.rad2deg(numpy.arctan2(self.col_vector.dot(shadow_prime), self.row_vector.dot(shadow_prime))),
            Magnitude=numpy.linalg.norm(shadow_prime))

    @property
    def Layover(self):
        # type: () -> AngleMagnitudeType
        """
        AngleMagnitudeType: The layover angle and magnitude.
        """

        layover = self.normal_vector - self.slant_z/(self.slant_z.dot(self.normal_vector))
        return AngleMagnitudeType(
            Angle=numpy.rad2deg(numpy.arctan2(self.col_vector.dot(layover), self.row_vector.dot(layover))),
            Magnitude=numpy.linalg.norm(layover))

    @property
    def North(self):
        """
        float: The north angle.
        """

        lat, lon, hae = ecf_to_geodetic(self.SCP)
        lat_r = numpy.deg2rad(lat)
        lon_r = numpy.deg2rad(lon)
        north = numpy.array(
            [-numpy.sin(lat_r)*numpy.cos(lon_r),
             -numpy.sin(lat_r)*numpy.sin(lon_r),
             numpy.cos(lat_r)])
        north_prime = north - self.slant_z*(north.dot(self.normal_vector)/self.slant_z.dot(self.normal_vector))
        return numpy.rad2deg(numpy.arctan2(self.col_vector.dot(north_prime), self.row_vector.dot(north_prime)))

    @property
    def MultiPath(self):
        """
        float: The multipath angle.
        """

        multipath = self.slant_x - self.slant_z*(
                self.slant_x.dot(self.normal_vector)/self.slant_z.dot(self.normal_vector))
        return numpy.rad2deg(numpy.arctan2(self.col_vector.dot(multipath), self.row_vector.dot(multipath)))

    @property
    def GroundTrack(self):
        """
        float: The ground track angle.
        """

        track = self.ARPVel - (self.ARPVel.dot(self.normal_vector))*self.normal_vector
        return numpy.rad2deg(numpy.arctan2(self.col_vector.dot(track), self.row_vector.dot(track)))

    def get_ground_plane_resolution(self, row_ss, col_ss):
        """
        Get the resolution in the ground plane.

        Parameters
        ----------
        row_ss : float
        col_ss : float

        Returns
        -------
        float, float
        """

        x_g = self.slant_x - (self.slant_x.dot(self.normal_vector))*self.normal_vector
        theta_r = -numpy.arctan2(self.col_vector.dot(x_g), self.row_vector.dot(x_g))
        graze = numpy.deg2rad(self.GrazeAngle)
        tilt = numpy.deg2rad(self.TiltAngle)

        k_r1 = (numpy.cos(theta_r)/numpy.cos(graze))**2 + \
               (numpy.sin(theta_r)**2*numpy.tan(graze)*numpy.tan(tilt) -
                numpy.sin(2*theta_r)/numpy.cos(graze))*numpy.tan(graze)*numpy.tan(tilt)
        k_r2 = (numpy.sin(theta_r)/numpy.cos(tilt))**2

        k_c1 = (numpy.sin(theta_r)**2/numpy.cos(graze) -
                numpy.sin(2*theta_r)*numpy.tan(graze)*numpy.tan(tilt))/numpy.cos(graze) + \
               (numpy.cos(theta_r)*numpy.tan(graze)*numpy.tan(tilt))**2
        k_c2 = (numpy.cos(theta_r)/numpy.cos(tilt))**2

        r2 = row_ss*row_ss
        c2 = col_ss*col_ss
        return float(numpy.sqrt(k_r1*r2 + k_r2*c2)), float(numpy.sqrt(k_c1*r2 + k_c2*c2))

    @classmethod
    def from_sicd(cls, sicd, row_vector, col_vector):
        """
        Construct from SICD structure.

        Parameters
        ----------
        sicd : SICDType
        row_vector : numpy.ndarray
        col_vector : numpy.ndarray
        """

        calculator = GeometryCalculator(
            sicd.GeoData.SCP.ECF.get_array(),
            sicd.SCPCOA.ARPPos.get_array(),
            sicd.SCPCOA.ARPVel.get_array())
        return cls(calculator, row_vector, col_vector)


def _extract_sicd_tx_rcv_pol(str_in):
    """
    Extract the tx and rcv components from the sicd style tx/rcv polarization string.

    Parameters
    ----------
    str_in : str

    Returns
    -------
    str, str
    """

    if str_in is None:
        return 'UNKNOWN', 'UNKNOWN'

    if not isinstance(str_in, str):
        raise TypeError('requires a string type input.')

    if str_in in ['OTHER', 'UNKNOWN']:
        return str_in, str_in

    def _decode_single_pol(pol):
        if pol in RCV_POLARIZATION_VALUES:
            return pol
        if pol in ['X', 'Y', 'S', 'E'] or pol.startswith('OTHER'):
            return 'OTHER'
        return 'UNKNOWN'

    tx, rcv = str_in.split(':')
    tx = _decode_single_pol(tx)
    rcv = _decode_single_pol(rcv)

    return tx, rcv


class InputROIType(Serializable):
    """
    ROI representing portion of input data used to make this product.
    """

    _fields = ('Size', 'UpperLeft')
    _required = ('Size', 'UpperLeft')
    # Descriptor
    Size = SerializableDescriptor(
        'Size', RowColIntType, _required, strict=DEFAULT_STRICT,
        docstring='Number of rows and columns extracted from the input.')  # type: RowColIntType
    UpperLeft = SerializableDescriptor(
        'UpperLeft', RowColIntType, _required, strict=DEFAULT_STRICT,
        docstring='The upper-left pixel extracted from the input.')  # type: RowColIntType

    def __init__(self, Size=None, UpperLeft=None, **kwargs):
        """

        Parameters
        ----------
        Size : RowColIntType|numpy.ndarray|list|tuple
        UpperLeft : RowColIntType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Size = Size
        self.UpperLeft = UpperLeft
        super(InputROIType, self).__init__(**kwargs)


class TxRcvPolarizationType(Serializable):
    """
    The transmit/receive polarization information.
    """

    _fields = ('TxPolarization', 'RcvPolarization', 'RcvPolarizationOffset')
    _required = ('TxPolarization', 'RcvPolarization')
    _numeric_format = {'RcvPolarizationOffset': FLOAT_FORMAT}
    # Descriptor
    TxPolarization = StringEnumDescriptor(
        'TxPolarization', TX_POLARIZATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Transmit polarization type.')  # type: str
    RcvPolarization = StringEnumDescriptor(
        'RcvPolarization', RCV_POLARIZATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Receive polarization type.')  # type: str
    RcvPolarizationOffset = FloatModularDescriptor(
        'RcvPolarizationOffset', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle offset for the receive polarization defined at aperture center.')  # type: float

    def __init__(self, TxPolarization=None, RcvPolarization=None, RcvPolarizationOffset=None, **kwargs):
        """

        Parameters
        ----------
        TxPolarization : str
        RcvPolarization : str
        RcvPolarizationOffset : None|float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxPolarization = TxPolarization
        self.RcvPolarization = RcvPolarization
        self.RcvPolarizationOffset = RcvPolarizationOffset
        super(TxRcvPolarizationType, self).__init__(**kwargs)

    @classmethod
    def from_sicd_value(cls, str_in):
        """
        Construct from the sicd style tx/rcv polarization string.

        Parameters
        ----------
        str_in : str

        Returns
        -------
        TxRcvPolarizationType
        """

        tx, rcv = _extract_sicd_tx_rcv_pol(str_in)
        return cls(TxPolarization=tx, RcvPolarization=rcv)


class ProcTxRcvPolarizationType(Serializable):
    """
    The processed transmit/receive polarization.
    """
    _fields = ('TxPolarizationProc', 'RcvPolarizationProc')
    _required = ('TxPolarizationProc', 'RcvPolarizationProc')
    # Descriptor
    TxPolarizationProc = StringEnumDescriptor(
        'TxPolarizationProc', TX_POLARIZATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Transmit polarization type.')  # type: str
    RcvPolarizationProc = StringEnumDescriptor(
        'RcvPolarizationProc', RCV_POLARIZATION_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Receive polarization type.')  # type: str

    def __init__(self, TxPolarizationProc=None, RcvPolarizationProc=None, **kwargs):
        """

        Parameters
        ----------
        TxPolarizationProc : str
        RcvPolarizationProc : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxPolarizationProc = TxPolarizationProc
        self.RcvPolarizationProc = RcvPolarizationProc
        super(ProcTxRcvPolarizationType, self).__init__(**kwargs)

    @classmethod
    def from_sicd_value(cls, str_in):
        """
        Construct from the sicd style tx/rcv polarization string.

        Parameters
        ----------
        str_in : str

        Returns
        -------
        ProcTxRcvPolarizationType
        """

        tx, rcv = _extract_sicd_tx_rcv_pol(str_in)
        return cls(TxPolarizationProc=tx, RcvPolarizationProc=rcv)


class ExploitationFeaturesCollectionInformationType(Serializable):
    """
    General collection information.
    """

    _fields = (
        'SensorName', 'RadarMode', 'CollectionDateTime', 'LocalDateTime', 'CollectionDuration',
        'Resolution', 'InputROI', 'Polarizations')
    _required = ('SensorName', 'RadarMode', 'CollectionDateTime', 'CollectionDuration')
    _collections_tags = {'Polarizations': {'array': False, 'child_tag': 'Polarization'}}
    _numeric_format = {'CollectionDuration': FLOAT_FORMAT}
    # Descriptor
    SensorName = StringDescriptor(
        'SensorName', _required, strict=DEFAULT_STRICT,
        docstring='The name of the sensor.')  # str
    RadarMode = SerializableDescriptor(
        'RadarMode', RadarModeType, _required, strict=DEFAULT_STRICT,
        docstring='Radar collection mode.')  # type: RadarModeType
    CollectionDateTime = DateTimeDescriptor(
        'CollectionDateTime', _required, strict=DEFAULT_STRICT,
        docstring='Collection date and time defined in Coordinated Universal Time (UTC). The seconds '
                  'should be followed by a Z to indicate UTC.')  # type: numpy.datetime64
    CollectionDuration = FloatDescriptor(
        'CollectionDuration', _required, strict=DEFAULT_STRICT,
        docstring='The duration of the collection (units = seconds).')  # type: float
    Resolution = SerializableDescriptor(
        'Resolution', RangeAzimuthType, _required, strict=DEFAULT_STRICT,
        docstring='Uniformly-weighted resolution (range and azimuth) processed in '
                  'the slant plane.')  # type: Union[None, RangeAzimuthType]
    InputROI = SerializableDescriptor(
        'InputROI', InputROIType, _required, strict=DEFAULT_STRICT,
        docstring='ROI representing portion of input data used to make '
                  'this product.')  # type: Union[None, InputROIType]
    Polarizations = SerializableListDescriptor(
        'Polarizations', TxRcvPolarizationType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Transmit and receive polarization(s).')  # type: Union[None, List[TxRcvPolarizationType]]

    def __init__(self, SensorName=None, RadarMode=None, CollectionDateTime=None, LocalDateTime=None,
                 CollectionDuration=None, Resolution=None, Polarizations=None, **kwargs):
        """

        Parameters
        ----------
        SensorName : str
        RadarMode : RadarModeType
        CollectionDateTime : numpy.datetime64|datetime.datetime|datetime.date|str
        LocalDateTime : None|str|datetime.datetime
        CollectionDuration : float
        Resolution : None|RangeAzimuthType|numpy.ndarray|list|tuple
        Polarizations : None|List[TxRcvPolarizationType]
        kwargs
        """

        self._local_date_time = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SensorName = SensorName
        self.RadarMode = RadarMode
        self.CollectionDateTime = CollectionDateTime
        self.CollectionDuration = CollectionDuration
        self.LocalDateTime = LocalDateTime
        self.Resolution = Resolution
        self.Polarizations = Polarizations
        super(ExploitationFeaturesCollectionInformationType, self).__init__(**kwargs)

    @property
    def LocalDateTime(self):
        """None|str:  The local date/time string of the collection. *Optional.*"""
        return self._local_date_time

    @LocalDateTime.setter
    def LocalDateTime(self, value):  # type: (Union[None, str, datetime.datetime]) -> None
        if value is None:
            self._local_date_time = None
            return
        elif isinstance(value, datetime.datetime):
            value = value.isoformat('T')

        if isinstance(value, str):
            self._local_date_time = value
        else:
            logger.error(
                'Attribute LocalDateTime of class ExploitationFeaturesCollectionInformationType\n\t'
                'requires a datetime.datetime or string. Got unsupported type {}.\n\t'
                'Setting value to None.'.format(type(value)))
            self._local_date_time = None

    @classmethod
    def from_sicd(cls, sicd):
        """
        Construct from a sicd element.

        Parameters
        ----------
        sicd : SICDType

        Returns
        -------
        ExploitationFeaturesCollectionInformationType
        """

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))

        polarizations = [
            TxRcvPolarizationType.from_sicd_value(entry.TxRcvPolarization)
            for entry in sicd.RadarCollection.RcvChannels]

        return cls(SensorName=sicd.CollectionInfo.CollectorName,
                   RadarMode=RadarModeType(**sicd.CollectionInfo.RadarMode.to_dict()),
                   CollectionDateTime=sicd.Timeline.CollectStart,
                   CollectionDuration=sicd.Timeline.CollectDuration,
                   Resolution=get_collection_resolution(sicd),
                   Polarizations=polarizations)


class ExploitationFeaturesCollectionGeometryType(Serializable):
    """
    Key geometry parameters independent of product processing. All values
    computed at the center time of the full collection.
    """

    _fields = ('Azimuth', 'Slope', 'Squint', 'Graze', 'Tilt', 'DopplerConeAngle', 'Extensions')
    _required = ()
    _collections_tags = {'Extensions': {'array': False, 'child_tag': 'Extension'}}
    _numeric_format = {
        'Azimuth': FLOAT_FORMAT, 'Slope': FLOAT_FORMAT, 'Squint': FLOAT_FORMAT, 'Graze': FLOAT_FORMAT,
        'Tilt': FLOAT_FORMAT, 'DopplerConeAngle': FLOAT_FORMAT}
    # Descriptor
    Azimuth = FloatDescriptor(
        'Azimuth', _required, strict=DEFAULT_STRICT, bounds=(0.0, 360.0),
        docstring='Angle clockwise from north indicating the ETP line of sight vector.')  # type: float
    Slope = FloatDescriptor(
        'Slope', _required, strict=DEFAULT_STRICT, bounds=(0.0, 90.0),
        docstring='Angle between the ETP at scene center and the range vector perpendicular to '
                  'the direction of motion.')  # type: float
    Squint = FloatModularDescriptor(
        'Squint', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle from the ground track to platform velocity vector at nadir. '
                  'Left-look is positive, right-look is negative.')  # type: float
    Graze = FloatDescriptor(
        'Graze', _required, strict=DEFAULT_STRICT, bounds=(0.0, 90.0),
        docstring='Angle between the ETP and the line of sight vector.')  # type: float
    Tilt = FloatModularDescriptor(
        'Tilt', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Angle between the ETP and the cross range vector. '
                  'Also known as the twist angle.')  # type: float
    DopplerConeAngle = FloatDescriptor(
        'DopplerConeAngle', _required, strict=DEFAULT_STRICT, bounds=(0.0, 180.0),
        docstring='The angle between the velocity vector and the radar line-of-sight vector. '
                  'Also known as the slant plane squint angle.')  # type: float
    Extensions = ParametersDescriptor(
        'Extensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Exploitation feature extension related to geometry for a '
                  'single input image.')  # type: ParametersCollection

    def __init__(self, Azimuth=None, Slope=None, Squint=None, Graze=None, Tilt=None,
                 DopplerConeAngle=None, Extensions=None, **kwargs):
        """

        Parameters
        ----------
        Azimuth : None|float
        Slope : None|float
        Squint : None|float
        Graze : None|float
        Tilt : None|float
        DopplerConeAngle : None|float
        Extensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Azimuth = Azimuth
        self.Slope = Slope
        self.Squint = Squint
        self.Graze = Graze
        self.Tilt = Tilt
        self.DopplerConeAngle = DopplerConeAngle
        self.Extensions = Extensions
        super(ExploitationFeaturesCollectionGeometryType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator):
        """
        Create from an ExploitationCalculator object.

        Parameters
        ----------
        calculator : ExploitationCalculator

        Returns
        -------
        ExploitationFeaturesCollectionGeometryType
        """

        if not isinstance(calculator, ExploitationCalculator):
            raise TypeError(_exp_calc_text.format(type(calculator)))

        return cls(Azimuth=calculator.AzimuthAngle,
                   Slope=calculator.SlopeAngle,
                   Graze=calculator.GrazeAngle,
                   Tilt=calculator.TiltAngle,
                   DopplerConeAngle=calculator.DopplerConeAngle,
                   Squint=calculator.SquintAngle)


class ExploitationFeaturesCollectionPhenomenologyType(Serializable):
    """
    Phenomenology related to both the geometry and the final product processing.
    All values computed at the center time of the full collection.
    """

    _fields = ('Shadow', 'Layover', 'MultiPath', 'GroundTrack', 'Extensions')
    _required = ()
    _collections_tags = {'Extensions': {'array': False, 'child_tag': 'Extension'}}
    _numeric_format = {'MultiPath': FLOAT_FORMAT, 'GroundTrack': FLOAT_FORMAT}
    # Descriptor
    Shadow = SerializableDescriptor(
        'Shadow', AngleMagnitudeType, _required, strict=DEFAULT_STRICT,
        docstring='The phenomenon where vertical objects occlude radar '
                  'energy.')  # type: Union[None, AngleMagnitudeType]
    Layover = SerializableDescriptor(
        'Layover', AngleMagnitudeType, _required, strict=DEFAULT_STRICT,
        docstring='The phenomenon where vertical objects appear as ground objects with '
                  'the same range/range rate.')  # type: Union[None, AngleMagnitudeType]
    MultiPath = FloatModularDescriptor(
        'MultiPath', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='This is a range dependent phenomenon which describes the energy from a '
                  'single scatter returned to the radar via more than one path and results '
                  'in a nominally constant direction in the ETP.')  # type: Union[None, float]
    GroundTrack = FloatModularDescriptor(
        'GroundTrack', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Counter-clockwise angle from increasing row direction to ground track '
                  'at the center of the image.')  # type: Union[None, float]
    Extensions = ParametersDescriptor(
        'Extensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Exploitation feature extension related to geometry for a '
                  'single input image.')  # type: ParametersCollection

    def __init__(self, Shadow=None, Layover=None, MultiPath=None, GroundTrack=None, Extensions=None, **kwargs):
        """

        Parameters
        ----------
        Shadow : None|AngleMagnitudeType|numpy.ndarray|list|tuple
        Layover : None|AngleMagnitudeType|numpy.ndarray|list|tuple
        MultiPath : None|float
        GroundTrack : None|float
        Extensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Shadow = Shadow
        self.Layover = Layover
        self.MultiPath = MultiPath
        self.GroundTrack = GroundTrack
        self.Extensions = Extensions
        super(ExploitationFeaturesCollectionPhenomenologyType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator):
        """
        Create from an ExploitationCalculator object.

        Parameters
        ----------
        calculator : ExploitationCalculator

        Returns
        -------
        ExploitationFeaturesCollectionPhenomenologyType
        """

        if not isinstance(calculator, ExploitationCalculator):
            raise TypeError(_exp_calc_text.format(type(calculator)))
        return cls(Shadow=calculator.Shadow, Layover=calculator.Layover,
                   MultiPath=calculator.MultiPath, GroundTrack=calculator.GroundTrack)


class CollectionType(Serializable):
    """
    Metadata regarding one of the input collections.
    """
    _fields = ('Information', 'Geometry', 'Phenomenology', 'identifier')
    _required = ('Information', 'identifier')
    _set_as_attribute = ('identifier', )
    # Descriptor
    Information = SerializableDescriptor(
        'Information', ExploitationFeaturesCollectionInformationType, _required, strict=DEFAULT_STRICT,
        docstring='General collection information.')  # type: ExploitationFeaturesCollectionInformationType
    Geometry = SerializableDescriptor(
        'Geometry', ExploitationFeaturesCollectionGeometryType, _required, strict=DEFAULT_STRICT,
        docstring='Key geometry parameters independent of product '
                  'processing.')  # type: Union[None, ExploitationFeaturesCollectionGeometryType]
    Phenomenology = SerializableDescriptor(
        'Phenomenology', ExploitationFeaturesCollectionPhenomenologyType, _required, strict=DEFAULT_STRICT,
        docstring='Phenomenology related to both the geometry and the final '
                  'product processing.')  # type: Union[None, ExploitationFeaturesCollectionPhenomenologyType]
    identifier = StringDescriptor(
        'identifier', _required, strict=DEFAULT_STRICT,
        docstring='The exploitation feature identifier.')  # type: str

    def __init__(self, Information=None, Geometry=None, Phenomenology=None, identifier=None, **kwargs):
        """

        Parameters
        ----------
        Information : ExploitationFeaturesCollectionInformationType
        Geometry : None|ExploitationFeaturesCollectionGeometryType
        Phenomenology : None|ExploitationFeaturesCollectionPhenomenologyType
        identifier : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Information = Information
        self.Geometry = Geometry
        self.Phenomenology = Phenomenology
        self.identifier = identifier
        super(CollectionType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator, sicd):
        """
        Create from an ExploitationCalculator object.

        Parameters
        ----------
        calculator : ExploitationCalculator
        sicd : SICDType

        Returns
        -------
        CollectionType
        """

        if not isinstance(calculator, ExploitationCalculator):
            raise TypeError(_exp_calc_text.format(type(calculator)))

        return cls(identifier=sicd.CollectionInfo.CoreName,
                   Information=ExploitationFeaturesCollectionInformationType.from_sicd(sicd),
                   Geometry=ExploitationFeaturesCollectionGeometryType.from_calculator(calculator),
                   Phenomenology=ExploitationFeaturesCollectionPhenomenologyType.from_calculator(calculator))


class ExploitationFeaturesProductType(Serializable):
    """
    Metadata regarding the product.
    """

    _fields = ('Resolution', 'Ellipticity', 'Polarizations', 'North', 'Extensions')
    _required = ('Resolution', 'Ellipticity', 'Polarizations')
    _collections_tags = {
        'Polarizations': {'array': False, 'child_tag': 'Polarization'},
        'Extensions': {'array': False, 'child_tag': 'Extension'}}
    _numeric_format = {'Ellipticity': FLOAT_FORMAT, 'North': FLOAT_FORMAT}
    # Descriptor
    Resolution = SerializableDescriptor(
        'Resolution', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='Uniformly-weighted resolution projected into the Earth Tangent '
                  'Plane (ETP).')  # type: RowColDoubleType
    Ellipticity = FloatDescriptor(
        'Ellipticity', _required, strict=DEFAULT_STRICT,
        docstring="Ellipticity of the 2D-IPR at the ORP, measured in the *Earth Geodetic "
                  "Tangent Plane (EGTP)*. Ellipticity is the ratio of the IPR ellipse's "
                  "major axis to minor axis.")  # type: float
    Polarizations = SerializableListDescriptor(
        'Polarizations', ProcTxRcvPolarizationType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Describes the processed transmit and receive polarizations for the '
                  'product.')  # type: List[ProcTxRcvPolarizationType]
    North = FloatModularDescriptor(
        'North', 180.0, _required, strict=DEFAULT_STRICT,
        docstring='Counter-clockwise angle from increasing row direction to north at the center '
                  'of the image.')  # type: float
    Extensions = ParametersDescriptor(
        'Extensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Exploitation feature extension related to geometry for a '
                  'single input image.')  # type: ParametersCollection

    def __init__(self, Resolution=None, Ellipticity=None, Polarizations=None,
                 North=None, Extensions=None, **kwargs):
        """

        Parameters
        ----------
        Resolution : RowColDoubleType|numpy.ndarray|list|tuple
        Ellipticity : float
        Polarizations : List[ProcTxRcvPolarizationType]
        North : None|float
        Extensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Resolution = Resolution
        self.Ellipticity = Ellipticity
        self.Polarizations = Polarizations
        self.North = North
        self.Extensions = Extensions
        super(ExploitationFeaturesProductType, self).__init__(**kwargs)

    @classmethod
    def from_calculator(cls, calculator, sicd):
        """
        Construct from a sicd element.

        Parameters
        ----------
        calculator : ExploitationCalculator
        sicd : SICDType

        Returns
        -------
        ExploitationFeaturesProductType
        """

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))

        rho_rg_slant, rho_az_slant = get_collection_resolution(sicd)
        rho_row_etp, rho_col_etp = calculator.get_ground_plane_resolution(rho_rg_slant, rho_az_slant)
        ellipticity = rho_row_etp/rho_col_etp if rho_row_etp >= rho_col_etp else rho_col_etp/rho_row_etp

        return cls(Resolution=(rho_row_etp, rho_col_etp),
                   Ellipticity=ellipticity,
                   Polarizations=[
                       ProcTxRcvPolarizationType.from_sicd_value(sicd.ImageFormation.TxRcvPolarizationProc), ],
                   North=calculator.North)


class ExploitationFeaturesType(Serializable):
    """
    Computed metadata regarding the collect.
    """
    _fields = ('Collections', 'Products')
    _required = ('Collections', 'Products')
    _collections_tags = {
        'Collections': {'array': False, 'child_tag': 'Collection'},
        'Products': {'array': False, 'child_tag': 'Product'}}
    # Descriptor
    Collections = SerializableListDescriptor(
        'Collections', CollectionType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[CollectionType]
    Products = SerializableListDescriptor(
        'Products', ExploitationFeaturesProductType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: List[ExploitationFeaturesProductType]

    def __init__(self, Collections=None, Products=None, **kwargs):
        """

        Parameters
        ----------
        Collections : List[CollectionType]
        Products : List[ExploitationFeaturesProductType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Collections = Collections
        self.Products = Products
        super(ExploitationFeaturesType, self).__init__(**kwargs)

    @classmethod
    def from_sicd(cls, sicd, row_vector, col_vector):
        """
        Construct from a sicd element.

        Parameters
        ----------
        sicd : SICDType|List[SICDType]
        row_vector : numpy.ndarray
        col_vector : numpy.ndarray

        Returns
        -------
        ExploitationFeaturesType
        """
        if isinstance(sicd, (list, tuple)):
            collections = []
            feats = []
            for i, entry in sicd:
                calculator = ExploitationCalculator.from_sicd(entry, row_vector, col_vector)
                collections.append(CollectionType.from_calculator(calculator, entry))
                feats.append(ExploitationFeaturesProductType.from_calculator(calculator, entry))
            return cls(Collections=collections, Products=feats)

        if not isinstance(sicd, SICDType):
            raise TypeError(_sicd_type_text.format(type(sicd)))
        calculator = ExploitationCalculator.from_sicd(sicd, row_vector, col_vector)
        return cls(
            Collections=[CollectionType.from_calculator(calculator, sicd), ],
            Products=[ExploitationFeaturesProductType.from_calculator(calculator, sicd)])


def get_collection_resolution(sicd_meta):
    """Compute the uniformly-weighted resolution (half-power width) in the slant plane from SICD metadata"""
    hpbw_rectangular = sarpy.processing.sicd.windows.find_half_power(numpy.ones(2**10))

    delta_xrow = hpbw_rectangular / sicd_meta.Grid.Row.ImpRespBW
    delta_ycol = hpbw_rectangular / sicd_meta.Grid.Col.ImpRespBW
    m_spxy_il = sarpy.geometry.point_projection.image_to_slant_sensitivity(sicd_meta, delta_xrow, delta_ycol)

    delta_spx, delta_spy = m_spxy_il * [delta_xrow, delta_ycol]
    return max(abs(delta_spx)), max(abs(delta_spy))



"""
Create kmz products based on SAR products.

For a basic help on the command-line, check

>>> python -m sarpy.utils.create_kmz --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import argparse
import logging
import os

import sarpy
from sarpy.io.complex.converter import open_complex
import sarpy.io.general.base
import sarpy.io.phase_history
import sarpy.io.received.converter
from sarpy.visualization.kmz_product_creation import create_kmz_view
from sarpy.visualization.cphd_kmz_product_creation import cphd_create_kmz_view
from sarpy.visualization.crsd_kmz_product_creation import crsd_create_kmz_view


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="Create KMZ product from a CRSD, CPHD or Complex Image.",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        'input_file', metavar='input_file',
        help='Path input data file, or directory for radarsat, RCM, or sentinel.\n'
             '* For radarsat or RCM, this can be the product.xml file, or parent directory\n'
             '  of product.xml or metadata/product.xml.\n'
             '* For sentinel, this can be the manifest.safe file, or parent directory of\n'
             '  manifest.safe.\n')
    parser.add_argument(
        'output_directory', metavar='output_directory',
        help='Path to the output directory where the product file(s) will be created.\n'
             'This directory MUST exist.\n'
             '* Depending on the input file, multiple product files may be produced.\n'
             '* The name for the output file(s) will be chosen based on CoreName and\n '
             '  transmit/collect polarization.\n')
    parser.add_argument(
        '-s', '--size', default=3072, type=int,
        help='(complex image only) Maximum # of thumbnail rows/columns, put -1 for full size')
    parser.add_argument(
        '--include-all', action='store_true', help='Include as much as possible in the KMZ')
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose (level="INFO") logging?')

    args = parser.parse_args()

    level = 'INFO' if args.verbose else 'WARNING'
    logging.basicConfig(level=level)
    logger = logging.getLogger('sarpy')
    logger.setLevel(level)

    file_stem = 'View-' + os.path.splitext(os.path.split(args.input_file)[1])[0]

    def _cphd_kmz():
        reader = sarpy.io.phase_history.open(args.input_file)
        cphd_create_kmz_view(reader, args.output_directory, file_stem=file_stem)

    def _crsd_kmz():
        reader = sarpy.io.received.converter.open_received(args.input_file)
        crsd_create_kmz_view(reader, args.output_directory, file_stem=file_stem)

    def _complex_image_kmz():
        reader = open_complex(args.input_file)
        pixel_limit = None if args.size == -1 else args.size
        create_kmz_view(reader, args.output_directory, pixel_limit=pixel_limit, file_stem=file_stem,
                        inc_image_corners=args.include_all,
                        inc_valid_data=args.include_all,
                        inc_collection_wedge=args.include_all,
                        inc_antenna=args.include_all)

    for func in (_cphd_kmz, _crsd_kmz, _complex_image_kmz):
        try:
            func()
            break
        except sarpy.io.general.base.SarpyIOError:
            continue
    else:
        raise sarpy.io.general.base.SarpyIOError(f'A KMZ generator for {args.input_file} could not be found')



"""
Classes and methods for parsing and using digital elevation models (DEM) in GeoTIFF format.

This code makes the following assumptions.
    1. The GeoTIFF files tile the earth with one degree offsets in both latitude and longitude.
    2. There is one pixel of overlap between adjacent tiles.
    3. The south-west corner of each tile is at an integer (degrees) latitude and longitude.
    4. The latitude and longitude of south-west corner points is encoded in the GeoTIFF filename.
    5. The anti-meridian is at W180 rather than at E180 so that valid longitude values are (-180 <= lon < 180) degrees.
"""
import glob
import logging
import pathlib
import warnings

import numpy as np
try:
    from PIL import Image
    from PIL import TiffTags
    Image.MAX_IMAGE_PIXELS = None  # get rid of decompression bomb checking
except ImportError:
    Image = None
    TiffTags = None

from scipy.interpolate import RegularGridInterpolator

from sarpy.io.DEM.DEM import DEMList
from sarpy.io.DEM.DEM import DEMInterpolator
from sarpy.io.DEM.geoid import GeoidHeight

logger = logging.getLogger(__name__)

__classification__ = "UNCLASSIFIED"
__author__ = "Valkyrie Systems Corporation"



class GeoTIFF1DegReader:
    """Class to read in a GeoTIFF file, if necessary, and cache the data."""

    def __init__(self, filename):
        self._filename = filename
        self._dem_data = None
        self._tiff_tags = None

    @property
    def filename(self):
        return self._filename

    @property
    def dem_data(self):
        if self._dem_data is None:
            self._read()  # pragma no cover
        return self._dem_data

    @property
    def tiff_tags(self):
        if self._tiff_tags is None:
            self._read()  # pragma no cover
        return self._tiff_tags

    def _read(self):
        # Note: the dem_data must have dtype=np.float64 otherwise the interpolator
        # created by RegularGridInterpolator will raise a TypeError exception.
        if Image is None or TiffTags is None:
            raise ImportError("Reading GeoTIFF DEM requires the PIL library")

        with Image.open(self._filename) as img:
            self._tiff_tags = {TiffTags.TAGS[key]: val for key, val in img.tag.items()}
            self._dem_data = np.asarray(img, dtype=np.float64)


class GeoTIFF1DegInterpolator(DEMInterpolator):
    """
    This class contains methods used to read DEM data from GeoTIFF files and interpolate the height values, as needed.

    Args
    ----
    dem_filename_pattern : str
        This is a format string that provides a glob pattern that will uniquely specify a DEM file from
        the Lat/Lon of the SW corner of the DEM tile.  See the GeoTIFF1DegList docstring for more details.
    ref_surface: str (default: "EGM2008")
        A case-insensitive string specifying the DEM reference surface. (eg., "WGS84" | "EGM2008" | "EGM96" | "EGM84")
    geoid_path: str | pathlib.Path | None (default: None)
        Optional filename of a specific Geoid file or a directory containing geoid files to choose from.
        If a directory is specified, then one or more of the following geoid files (in order of preference)
        will be chosen from this directory.

            'egm2008-1.pgm', 'egm2008-2_5.pgm', 'egm2008-5.pgm',
            'egm96-5.pgm', 'egm96-15.pgm', 'egm84-15.pgm', 'egm84-30.pgm'
    missing_error: bool (default: False)
        Optional flag indicating whether an exception will be raised when missing DEM data files are encountered.
        If True then a ValueError will be raised when a needed DEM data file can not be found.
        If False then a DEM value of zero will be used when a needed DEM data file is not found.
    interp_method: str (default: 'linear')
        Optional interpolation method. Any scipy.interpolate.RegularGridInterpolator method is valid here.
    max_readers: init (default: 4)
        Optional maximum number of DEM file readers.  A DEM file reader will read a DEM file and cache the results.
        DEM file readers can use a lot of memory (~8 bytes x number-of-DEM-samples), but will make processing faster.

     """
    __slots__ = ('_geoid_path', '_interp_method', '_ref_surface', '_geotiff_list_obj',
                 '_bounding_box_cache', '_max_readers', '_readers')

    def __init__(self, dem_filename_pattern, ref_surface='EGM2008', geoid_path=None, *,
                 missing_error=False, interp_method="linear", max_readers=4):
        self._geoid_path = pathlib.Path(geoid_path) if geoid_path else None
        self._interp_method = str(interp_method)
        self._ref_surface = str(ref_surface).upper()
        self._geotiff_list_obj = GeoTIFF1DegList(dem_filename_pattern, missing_error=missing_error)
        self._bounding_box_cache = {}
        self._max_readers = max(1, int(max_readers))
        self._readers = []

        # get the geoid object - we prefer egm2008*.pgm files, but in reality, it makes very little difference.
        if self._geoid_path and self._geoid_path.is_file():
            self._geoid_obj = GeoidHeight(str(self._geoid_path))
        elif self._geoid_path and self._geoid_path.is_dir():
            search_files = ('egm2008-1.pgm', 'egm2008-2_5.pgm', 'egm2008-5.pgm',
                            'egm96-5.pgm', 'egm96-15.pgm', 'egm84-15.pgm', 'egm84-30.pgm')
            self._geoid_obj = GeoidHeight.from_directory(str(self._geoid_path), search_files=search_files)
        else:
            self._geoid_obj = None

    @property
    def interp_method(self):
        return self._interp_method

    @interp_method.setter
    def interp_method(self, val):
        self._interp_method = str(val)

    def _read_dem_file(self, filename):
        """
        Get the DEM values and TIFF tags from the reader cache, if possible,
        otherwise create a new reader object in the reader cache and
        return its DEM data and TIFF tags.
        """
        for rdr in self._readers:
            if filename == rdr.filename:
                reader = rdr
                break
        else:
            if len(self._readers) >= self._max_readers:
                self._readers.pop(0)

            reader = GeoTIFF1DegReader(filename)
            self._readers.append(reader)

        return reader.tiff_tags, reader.dem_data

    def get_elevation_native(self, lat, lon, block_size=None):
        """
        Get the elevation value relative to the DEM file's reference surface.

        Parameters
        ----------
        lat : numpy.ndarray | list | tuple | int | float
        lon : numpy.ndarray | list | tuple | int | float
        block_size : int | None (default: None)
            Block processing is not supported; this argument is present to maintain a common interface with
            the DEMInterpolator parent class.  A value other than None will result in a warning.

        Returns
        -------
        numpy.ndarray
            The elevation relative to the reference surface of the DEM.
        """
        if block_size is not None:
            warnings.warn("Block processing is not implemented.  Full size processing will be used.")  # pragma nocover

        lat = np.atleast_1d(lat)
        lon = np.atleast_1d(lon)

        if lat.shape != lon.shape:
            raise ValueError("The lat and lon arrays are not the same shape.")

        lat_lon_pairs = np.stack([lat.flatten(), lon.flatten()], axis=-1)
        unique_sw_corners = np.unique(np.floor(lat_lon_pairs), axis=0)

        # Get the list of filenames for the unique SW corners, if they exist.
        filename_info = []
        for sw_lat, sw_lon in unique_sw_corners:
            # Adding a fractional offset to the otherwise integer SW corner Lat/Lon values
            # will guarantee that no more than one filename will be found.
            files = self._geotiff_list_obj.find_dem_files(sw_lat + 0.1, sw_lon + 0.1)
            if files:
                filename_info.append({"filename": files[0], "sw_lat": sw_lat, "sw_lon": sw_lon})

        height = np.zeros(lat.size)
        for info in filename_info:
            filename = info["filename"]
            tile_sw_lat = info["sw_lat"]
            tile_sw_lon = info["sw_lon"]
            tile_ne_lat = tile_sw_lat + 1
            tile_ne_lon = tile_sw_lon + 1

            tiff_tags, dem_data = self._read_dem_file(filename)

            gpars = tiff_tags.get('GeoAsciiParamsTag', ('',))[0].upper()
            implied_ref_surface = ('EGM84' if any([p in gpars for p in ['EGM84', 'EGM 84', 'EGM-84']]) else
                                   'EGM96' if any([p in gpars for p in ['EGM96', 'EGM 96', 'EGM-96']]) else
                                   'EGM2008' if any([p in gpars for p in ['EGM2008', 'EGM 2008', 'EGM-2008']]) else
                                   'EGM2020' if any([p in gpars for p in ['EGM2020', 'EGM 2020', 'EGM-2020']]) else
                                   'WGS84' if any([p in gpars for p in ['WGS84', 'WGS 84', 'WGS-84']]) else
                                   'Unknown')
            if ((self._ref_surface.startswith('EGM') and implied_ref_surface.startswith('WGS')) or
                    (self._ref_surface.startswith('WGS') and implied_ref_surface.startswith('EGM'))):
                msg = (f"{filename}\n"
                       f"The GeoAsciiParamsTag tag implies that the reference surface is {implied_ref_surface},\n"
                       f"but the explicit reference surface was defined to be {self._ref_surface}.\n"
                       f"This might cause the elevation values to be calculated incorrectly.\n")
                logger.warning(msg)

            tile_num_lats, tile_num_lons = dem_data.shape
            tile_lats = np.linspace(tile_ne_lat, tile_sw_lat, tile_num_lats)
            tile_lons = np.linspace(tile_sw_lon, tile_ne_lon, tile_num_lons)

            # Old versions of scipy.interpolate require that axis samples be in strictly ascending order.
            # Unfortunately, the tile_lats are in strictly descending order.  To get the interpolator to
            # work regardless of package version, we will negate tile_lats and the lat-part of lat_lon_pairs.
            neg_tile_lats = -tile_lats
            neg_lat_lon_pairs = [(-lat, lon) for lat, lon in lat_lon_pairs]

            interp = RegularGridInterpolator((neg_tile_lats, tile_lons), dem_data, method=self._interp_method,
                                             bounds_error=False, fill_value=np.nan)

            interp_height = interp(neg_lat_lon_pairs)
            mask = np.logical_not(np.isnan(interp_height))
            height[mask] = interp_height[mask]

        return height.reshape(lat.shape)

    def get_elevation_hae(self, lat, lon, block_size=None):
        """
        Get the elevation value relative to the WGS84 ellipsoid.

        Parameters
        ----------
        lat : numpy.ndarray | list | tuple | int | float
        lon : numpy.ndarray | list | tuple | int | float
        block_size : int | None (default: None)
            Block processing is not supported; this argument is present to maintain a common interface with
            the DEMInterpolator parent class.  A value other than None will result in a warning.

        Returns
        -------
        numpy.ndarray
            The elevation relative to the ellipsoid
        """
        height_native = self.get_elevation_native(lat, lon, block_size=block_size)

        if self._ref_surface.startswith('WGS'):
            return height_native
        elif self._ref_surface.startswith("EGM"):
            if self._geoid_obj is None:
                raise ValueError("The geoid_dir parameter was not defined so geoid calculations are disabled.")

            return height_native + self._geoid_obj.get(lat, lon, block_size=block_size)
        else:
            raise ValueError(f"The reference surface is {self._ref_surface}, which is not supported")

    def get_elevation_geoid(self, lat, lon, block_size=None):
        """
        Get the elevation value relative to the geoid.

        Parameters
        ----------
        lat : numpy.ndarray | list | tuple | int | float
        lon : numpy.ndarray | list | tuple | int | float
        block_size : int | None (default: None)
            Block processing is not supported; this argument is present to maintain a common interface with
            the DEMInterpolator parent class.  A value other than None will result in a warning.

        Returns
        -------
        numpy.ndarray
            the elevation relative to the geoid
        """
        height_native = self.get_elevation_native(lat, lon, block_size)

        if self._ref_surface.startswith("EGM"):
            return height_native
        elif self._ref_surface.startswith('WGS'):
            if self._geoid_obj is None:
                raise ValueError("The geoid_dir parameter was not defined so geoid calculations are disabled.")

            return height_native - self._geoid_obj.get(lat, lon, block_size=block_size)
        else:
            raise ValueError(f"The reference surface is {self._ref_surface}, which is not supported.")

    def get_max_hae(self, lat_lon_box=None):
        """
        Get the maximum dem value with respect to the ellipsoid, which should be assumed **approximately** correct.

        Parameters
        ----------
        lat_lon_box : list | numpy.ndarray
            Any area of interest of the form `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        float
        """
        result = self.get_min_max_native(lat_lon_box)

        if self._ref_surface.startswith('WGS'):
            return result['max']['height']
        else:
            return self.get_elevation_hae(result['max']['lat'], result['max']['lon'])[0]

    def get_min_hae(self, lat_lon_box=None):
        """
        Get the minimum dem value with respect to the ellipsoid, which should be assumed **approximately** correct.

        Parameters
        ----------
        lat_lon_box : list | numpy.ndarray
            Any area of interest of the form `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        float
        """
        result = self.get_min_max_native(lat_lon_box)

        if self._ref_surface.startswith('WGS'):
            return result['min']['height']
        else:
            return self.get_elevation_hae(result['min']['lat'], result['min']['lon'])[0]

    def get_max_geoid(self, lat_lon_box=None):
        """
        Get the maximum dem value with respect to the geoid, which should be assumed **approximately** correct.

        Parameters
        ----------
        lat_lon_box : list | numpy.ndarray
            Any area of interest of the form `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        float
        """
        result = self.get_min_max_native(lat_lon_box)

        if self._ref_surface.startswith('EGM'):
            return result['max']['height']
        else:
            return self.get_elevation_geoid(result['max']['lat'], result['max']['lon'])[0]

    def get_min_geoid(self, lat_lon_box=None):
        """
        Get the minimum dem value with respect to geoid, which should be assumed **approximately** correct.

        Parameters
        ----------
        lat_lon_box : list | numpy.ndarray
            Any area of interest of the form `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        float
        """
        result = self.get_min_max_native(lat_lon_box)

        if self._ref_surface.startswith('EGM'):
            return result['min']['height']
        else:
            return self.get_elevation_geoid(result['min']['lat'], result['min']['lon'])[0]

    def get_min_max_native(self, lat_lon_box):
        """
        Get the minimum and maximum dem value with respect to the native reference surface of the DEM.

        Parameters
        ----------
        lat_lon_box : List | numpy.ndarray
            The bounding box to search `[lat min lat max, lon min, lon max]`.

        Returns
        -------
        dict
            A dictionary describing the results of the search::

                {"box": lat_lon_box,
                 "min": {"lat": lat_deg, "lon": lon_deg, "height": height},
                 "max": {"lat": lat_deg, "lon": lon_deg, "height": height}}

        """
        if np.array_equal(self._bounding_box_cache.get("box", []), lat_lon_box):
            # If we have already done this calculation then don't do it again.
            return self._bounding_box_cache

        box_lat_min, box_lat_max, box_lon_min, box_lon_max = lat_lon_box

        filename_info = []
        for sw_lat in np.arange(np.floor(box_lat_min), np.floor(box_lat_max) + 1):
            for sw_lon in np.arange(np.floor(box_lon_min), np.floor(box_lon_max) + 1):
                files = self._geotiff_list_obj.find_dem_files(sw_lat + 0.1, sw_lon + 0.1)
                if files:
                    filename_info.append({"filename": files[0], "sw_lat": sw_lat, "sw_lon": sw_lon})

        # Initialize so that the global min and max occur at the same lat/lon and have a value of zero.
        # These values will be returned if the bounding box is completely outside the available DEM tiles.
        global_min_lat = box_lat_min
        global_max_lat = box_lat_min
        global_min_lon = box_lon_min
        global_max_lon = box_lon_min
        global_min = np.inf if filename_info else 0
        global_max = -np.inf if filename_info else 0

        for info in filename_info:
            filename = info["filename"]
            tile_sw_lat = info["sw_lat"]
            tile_sw_lon = info["sw_lon"]
            tile_ne_lat = tile_sw_lat + 1

            tiff_tags, dem_data = self._read_dem_file(filename)

            tile_num_lats = tiff_tags['ImageLength'][0]
            tile_num_lons = tiff_tags['ImageWidth'][0]

            # Lat index is in descending order, so calculate the offset from the north edge (lowest index)
            lat_start_offset = max(0, tile_ne_lat - box_lat_max)
            lat_stop_offset = min(1, tile_ne_lat - box_lat_min)

            # Lon index is in ascending order, so calculate the offset from the west edge (lowest index)
            lon_start_offset = max(0, box_lon_min - tile_sw_lon)
            lon_stop_offset = min(1, box_lon_max - tile_sw_lon)

            # Lat index is in descending order, so start at the north edge (lowest index)
            row_start = int(np.ceil(lat_start_offset * (tile_num_lats - 1)))
            row_stop = int(np.floor(lat_stop_offset * (tile_num_lats - 1)))

            # Lon index is in ascending order, so start at the west edge (lowest index)
            col_start = int(np.ceil(lon_start_offset * (tile_num_lons - 1)))
            col_stop = int(np.floor(lon_stop_offset * (tile_num_lons - 1)))

            dem_slice = dem_data[row_start:row_stop + 1, col_start:col_stop + 1]
            max_index = np.unravel_index(np.argmax(dem_slice), shape=dem_slice.shape)
            min_index = np.unravel_index(np.argmin(dem_slice), shape=dem_slice.shape)

            if global_max < dem_slice[max_index]:
                global_max = dem_slice[max_index]
                global_max_lat = tile_ne_lat - lat_start_offset - max_index[0] / (tile_num_lats - 1)
                global_max_lon = tile_sw_lon + lon_start_offset + max_index[1] / (tile_num_lons - 1)

            if global_min > dem_slice[min_index]:
                global_min = dem_slice[min_index]
                global_min_lat = tile_ne_lat - lat_start_offset - min_index[0] / (tile_num_lats - 1)
                global_min_lon = tile_sw_lon + lon_start_offset + min_index[1] / (tile_num_lons - 1)

        self._bounding_box_cache = {"box": lat_lon_box,
                                    "min": {"lat": global_min_lat, "lon": global_min_lon, "height": float(global_min)},
                                    "max": {"lat": global_max_lat, "lon": global_max_lon, "height": float(global_max)}
                                    }

        return self._bounding_box_cache


# ---------------------------------------------------------------------------------------------------------------------
# GeoTIFF1DegList
# ---------------------------------------------------------------------------------------------------------------------
class GeoTIFF1DegList(DEMList):
    """
    GeoTIFF subclass of sarpy.io.DEM.DEMList

    This class contains methods used to determine which GeoTIFF files are needed to cover
    a specified geodetic bounding box.

    Args
    ----
    dem_filename_pattern : str
        This is a format string that specifies the glob pattern that will uniquely specify a DEM file from
        the Lat/Lon of the SW corner of the DEM tile.  See the note below for more details.
    missing_error: bool (default: False)
        Optional flag indicating whether an exception will be raised when missing DEM data files are encountered.
        If True then a ValueError will be raised when a needed DEM data file can not be found.
        If False then a DEM value of zero will be used when a needed DEM data file is not found.

    Notes
    -----
    The DEM files must have the SW corner Lat/Lon encoded in their filenames.
    The dem_filename_pattern argument contains a format string that, when populated,
    will create a glob pattern that will specify the desired DEM file.  The following
    arguments are provided to the format string::

        lat = int(numpy.floor(lat))
        lon = int(numpy.floor(lon))
        abslat = int(abs(numpy.floor(lat)))
        abslon = int(abs(numpy.floor(lon)))
        ns = 's' if lat < 0 else 'n'
        NS = 'S' if lat < 0 else 'N'
        ew = 'w' if lon < 0 else 'e'
        EW = 'W' if lon < 0 else 'E'

    An example (with Linux file separators):
        "/dem_root/tdt_{ns}{abslat:02}{ew}{abslon:03}_*/DEM/TDT_{NS}{abslat:02}{EW}{abslon:03}_*_DEM.tif"

    will match filenames like:
        "/dem_root/tdt_n45e013_02/DEM/TDT_N45E013_02_DEM.tif"
        "/dem_root/tdt_s09w140_01/DEM/TDT_S09W140_01_DEM.tif"

    """

    __slots__ = ('_dem_filename_pattern', '_missing_error', '_pattern_to_filename')

    def __init__(self, dem_filename_pattern, missing_error=False):
        self._dem_filename_pattern = str(dem_filename_pattern)
        self._missing_error = bool(missing_error)
        self._pattern_to_filename = {}

    @staticmethod
    def filename_from_lat_lon(lat, lon, pattern):
        """
        This method will return the filename glob of the GeoTIFF file that contains the specified latitude/longitude.

        """
        pars = {
            "lat": int(np.floor(lat)),
            "lon": int(np.floor(lon)),
            "abslat": int(abs(np.floor(lat))),
            "abslon": int(abs(np.floor(lon))),
            "ns": 's' if lat < 0 else 'n',
            "ew": 'w' if lon < 0 else 'e',
            "NS": 'S' if lat < 0 else 'N',
            "EW": 'W' if lon < 0 else 'E',
        }

        class SkipMissing(dict):
            def __missing__(self, key):
                return f'{{{key}}}'

        return pattern.format_map(SkipMissing(**pars))

    def find_dem_files(self, lat, lon):
        """
        Return a list of filenames of GeoTIFF files that contain DEM data for the specified Lat/Lon point.
        Since DEM files overlap, there might be more than one file that contains the specified Lat/Lon point.

        Args
        ----
        lat: int | float
            The latitude in degrees (-90 <= lat <= 90)
        lon: int | float
            The longitude in degrees (-180 <= lon < 180)

        Returns
        -------
        filenames: list(str)
            A list of filenames of DEM data files, if they exists, otherwise []

        """
        msg = [] if -90.0 <= lat <= 90.0 else ["The latitude value must be between [-90, +90]"]
        msg += [] if -180.0 <= lon < 180.0 else ["The longitude value must be between [-180, +180)"]
        if msg:
            raise ValueError('\n'.join(msg))

        sw_lats = [89 if lat == 90 else np.floor(lat)]  # The latitude of the south-west corner in integer degrees
        sw_lons = [np.floor(lon)]  # The longitude of the south-west corner in integer degrees

        if lat == np.floor(lat) and np.abs(lat) < 90:
            # lat is an integer, so it is in the overlap region of at least two files
            sw_lats.append(np.floor(lat) - 1)

        if lon == np.floor(lon):
            # lon is an integer, so it is in the overlap region of at least two files.
            sw_lons.append(179 if lon == -180 else np.floor(lon) - 1)

        filenames = []
        for sw_lat in sw_lats:
            for sw_lon in sw_lons:
                glob_pattern = self.filename_from_lat_lon(int(sw_lat), int(sw_lon), self._dem_filename_pattern)
                if glob_pattern not in self._pattern_to_filename:
                    for filename in glob.glob(glob_pattern):
                        if pathlib.Path(filename).is_file():
                            # The glob should not return more than one filename,
                            # but if it does then keep only the first.
                            self._pattern_to_filename[glob_pattern] = filename
                            break
                    else:
                        self._pattern_to_filename[glob_pattern] = None
                        msg = f'Missing expected DEM file for tile with lower left lat/lon corner ({sw_lat}, {sw_lon})'
                        if self._missing_error:
                            raise ValueError(msg)
                        else:
                            logger.warning(
                                msg + '\n\tThis should result in the assumption that the altitude in\n\t'
                                    'that section is zero relative to the reference surface.')
                if self._pattern_to_filename[glob_pattern] is not None:
                    filenames.append(self._pattern_to_filename[glob_pattern])

        return filenames

    def get_file_list(self, lat_lon_box):
        """
        This will return the list of files associated with covering the `lat_lon_box` using a DEM.

        If the bounding box spans the anti-meridian (180th meridian), then the maximum longitude
        will be less than the minimum longitude.

        Args
        ----
        lat_lon_box : numpy.ndarray | list | tuple
            The bounding box of the form `[lat min, lat max, lon min, lon max]` in degrees.

        Returns
        -------
        filenames: List[str]
            A list of filenames, without duplication, of the files needed to cover the bounding box.
        """
        filenames = []

        lat_min, lat_max, lon_min, lon_max = lat_lon_box

        msg = ["The minimum latitude value must be between [-90, +90]"] if not (-90.0 <= lat_min <= 90.0) else []
        msg += ["The maximum latitude value must be between [-90, +90]"] if not (-90.0 <= lat_max <= 90.0) else []
        msg += ["The minimum longitude value must be between [-180, +180)"] if not (-180.0 <= lon_min < 180.0) else []
        msg += ["The maximum longitude value must be between [-180, +180)"] if not (-180.0 <= lon_max < 180.0) else []
        if msg:
            raise ValueError('\n'.join(msg))

        if lon_max < lon_min:
            lon_max += 360

        for lat_inc in np.arange(np.ceil(lat_max) - np.floor(lat_min)):
            lat = lat_min + lat_inc

            for lon_inc in np.arange(np.ceil(lon_max) - np.floor(lon_min)):
                lon = (lon_min + lon_inc + 180) % 360 - 180

                new_filenames = self.find_dem_files(lat, lon)
                new_unique_filenames = [file for file in new_filenames if file not in filenames]
                filenames.extend(new_unique_filenames)

        return filenames




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class RECORD(TREElement):
    def __init__(self, value):
        super(RECORD, self).__init__()
        self.add_field('ENGLN', 'd', 2, value)
        self.add_field('ENGLBL', 's', self.ENGLN, value)
        self.add_field('ENGMTXC', 's', 4, value)
        self.add_field('ENGMTXR', 's', 4, value)
        self.add_field('ENGTYP', 's', 1, value)
        self.add_field('ENGDTS', 'd', 1, value)
        self.add_field('ENGDTU', 's', 2, value)
        self.add_field('ENGDATC', 'd', 8, value)
        self.add_field('ENGDATA', 'b', self.ENGDATC*self.ENGDTS, value)


class ENGRDAType(TREElement):
    def __init__(self, value):
        super(ENGRDAType, self).__init__()
        self.add_field('RESRC', 's', 20, value)
        self.add_field('RECNT', 'd', 3, value)
        self.add_loop('RECORDSs', self.RECNT, RECORD, value)


class ENGRDA(TREExtension):
    _tag_value = 'ENGRDA'
    _data_type = ENGRDAType



"""
The label extension subheader definitions - applies only to NITF version 2.0
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from .base import NITFElement, UserHeaderType, _IntegerDescriptor, _RawDescriptor, \
    _StringDescriptor, _StringEnumDescriptor, _NITFElementDescriptor
from .security import NITFSecurityTags0


class LabelSegmentHeader(NITFElement):
    """
    Symbol segment subheader for NITF version 2.0 - see standards document
    MIL-STD-2500A for more information.
    """

    _ordering = (
        'LA', 'LID', 'Security', 'ENCRYP', 'LFS', 'LCW', 'LCH',
        'LDLVL', 'LALVL', 'LLOC', 'LTC', 'LBC', 'UserHeader')
    _lengths = {
        'LA': 2, 'LID': 7, 'ENCRYP': 1, 'LFS': 1, 'LCW': 2, 'LCH': 2,
        'LDLVL': 3, 'LALVL': 3, 'LLOC': 10, 'LTC': 3, 'LBC': 3}
    #######
    LA = _StringEnumDescriptor(
        'LA', True, 2, {'LA', }, default_value='LA')  # type: str
    LID = _StringDescriptor('LID', True, 10)  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags0, default_args={})  # type: NITFSecurityTags0
    ENCRYP = _StringEnumDescriptor(
        'ENCRYP', True, 1, {'0'}, default_value='0',
        docstring='Encryption.')  # type: str
    LFS = _StringDescriptor('LFS', True, 1)  # type: str
    LCW = _StringDescriptor('LCW', True, 2, default_value='00')  # type: str
    LCH = _StringDescriptor('LCH', True, 2, default_value='00')  # type: str
    LDLVL = _IntegerDescriptor(
        'LDLVL', True, 3, default_value=1)  # type: int
    LALVL = _IntegerDescriptor(
        'LALVL', True, 3, default_value=1)  # type: int
    LLOC = _StringDescriptor('LLOC', True, 10)  # type: str
    LTC = _RawDescriptor(
        'LTC', True, 3, default_value=b'\x00\x00\x00')  # type: bytes
    LBC = _RawDescriptor(
        'LBC', True, 3, default_value=b'\xff\xff\xff')  # type: bytes
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={},
        docstring='User defined header.')  # type: UserHeaderType



"""
A module for performing a selection of validation checks on a SIDD (nitf) file,
or an xml file containing the sidd structure.

Use the `check_file` function directly, or perform using the command line

>>> python -m sarpy.consistency.sidd_consistency <file_name>

For more information, about command line usage, see

>>> python -m sarpy.consistency.sidd_consistency --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import sys
import argparse
import os
import re

from sarpy.consistency.sicd_consistency import check_sicd_data_extension
from sarpy.io.general.nitf import NITFDetails
from sarpy.io.general.nitf_elements.des import DataExtensionHeader, \
    DataExtensionHeader0
from sarpy.io.xml.base import parse_xml_from_string, validate_xml_from_string
from sarpy.io.product.sidd_schema import get_schema_path, get_urn_details, \
    get_specification_identifier

from sarpy.io.product.sidd3_elements.SIDD import SIDDType as SIDDType3
from sarpy.io.product.sidd2_elements.SIDD import SIDDType as SIDDType2
from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1


logger = logging.getLogger('validation')


def evaluate_xml_versus_schema(xml_string, urn_string):
    """
    Check validity of the xml string versus the appropriate schema.

    Parameters
    ----------
    xml_string : str|bytes
    urn_string : str

    Returns
    -------
    bool
    """

    # get schema path
    try:
        the_schema = get_schema_path(urn_string)
    except KeyError:
        logger.exception('SIDD: Failed finding the schema for urn {}'.format(urn_string))
        return False

    try:
        return validate_xml_from_string(xml_string, the_schema, output_logger=logger)
    except ImportError:
        return None


def _evaluate_xml_string_validity(xml_string=None):
    """
    Check the validity of the SIDD xml, as defined by the given string.

    Parameters
    ----------
    xml_string : str|bytes

    Returns
    -------
    is_valid : bool
    sidd_urn : str
    sidd : SIDDType1|SIDDType2|SIDDType3
    """

    root_node, xml_ns = parse_xml_from_string(xml_string)
    if 'default' not in xml_ns:
        raise ValueError(
            'Could not properly interpret the namespace collection from xml\n{}'.format(xml_ns))

    sidd_urn = xml_ns['default']
    # check that our urn is mapped
    try:
        _ = get_urn_details(sidd_urn)
        check_schema = True
    except Exception as e:
        logger.exception('SIDD: The SIDD namespace has unrecognized value')
        check_schema = False

    valid_xml = None
    if check_schema:
        valid_xml = evaluate_xml_versus_schema(xml_string, sidd_urn)
    if valid_xml is None:
        valid_xml = True

    # perform the various sidd structure checks
    if sidd_urn == 'urn:SIDD:1.0.0':
        the_sidd = SIDDType1.from_node(root_node, xml_ns=xml_ns)
        valid_sidd_contents = the_sidd.is_valid(recursive=True, stack=False)
    elif sidd_urn == 'urn:SIDD:2.0.0':
        the_sidd = SIDDType2.from_node(root_node, xml_ns=xml_ns)
        valid_sidd_contents = the_sidd.is_valid(recursive=True, stack=False)
    elif sidd_urn == 'urn:SIDD:3.0.0':
        the_sidd = SIDDType3.from_node(root_node, xml_ns=xml_ns)
        valid_sidd_contents = the_sidd.is_valid(recursive=True, stack=False)
    else:
        raise ValueError('Got unhandled urn {}'.format(sidd_urn))
    return valid_xml & valid_sidd_contents, sidd_urn, the_sidd


def check_sidd_data_extension(nitf_details, des_header, xml_string):
    """
    Evaluate a SIDD data extension for validity.

    Parameters
    ----------
    nitf_details : NITFDetails
    des_header : DataExtensionHeader|DataExtensionHeader0
    xml_string : str|bytes

    Returns
    -------
    is_valid : bool
    sidd : SIDDType1|SIDDType2|SIDDType3
    """

    def check_des_header_fields():
        # type: () -> bool
        des_id = des_header.DESID.strip() if nitf_details.nitf_version == '02.10' else des_header.DESTAG.strip()

        if des_id != 'XML_DATA_CONTENT':
            logger.warning('SIDD: Found old style SIDD DES Header. This is deprecated.')
            return True

        # make sure that the NITF urn is evaluated for sensibility
        nitf_urn = des_header.UserHeader.DESSHTN.strip()
        try:
            nitf_urn_details = get_urn_details(nitf_urn)
        except Exception:
            logger.exception('SIDD: The SIDD DES.DESSHTN must be a recognized urn')
            return False

        # make sure that the NITF urn and SIDD urn actually agree
        header_good = True
        if nitf_urn != xml_urn:
            logger.error('SIDD: The SIDD DES.DESSHTN ({}) and urn ({}) must agree'.format(nitf_urn, xml_urn))
            header_good = False

        # make sure that the NITF DES fields are populated appropriately for NITF urn
        if des_header.UserHeader.DESSHSI.strip() != get_specification_identifier():
            logger.error(
                'SIDD: DES.DESSHSI has value `{}`,\n\tbut should have value `{}`'.format(
                    des_header.UserHeader.DESSHSI.strip(), get_specification_identifier()))
            header_good = False

        nitf_version = nitf_urn_details['version']
        if des_header.UserHeader.DESSHSV.strip() != nitf_version:
            logger.error(
                'SIDD: DES.DESSHSV has value `{}`,\n\tbut should have value `{}` based on DES.DESSHTN `{}`'.format(
                    des_header.UserHeader.DESSHSV.strip(), nitf_version, nitf_urn))
            header_good = False

        nitf_date = nitf_urn_details['date']
        if des_header.UserHeader.DESSHSD.strip() != nitf_date:
            logger.warning(
                'SIDD: DES.DESSHSD has value `{}`,\n\tbut should have value `{}` based on DES.DESSHTN `{}`'.format(
                    des_header.UserHeader.DESSHSV.strip(), nitf_date, nitf_urn))
        return header_good

    def compare_sidd_class():
        # type: () -> bool

        if the_sidd.ProductCreation is None or the_sidd.ProductCreation.Classification is None or \
                the_sidd.ProductCreation.Classification.classification is None:
            logger.error(
                'SIDD: SIDD.ProductCreation.Classification.classification is not populated,\n\t'
                'so can not be compared with SIDD DES.DESCLAS `{}`'.format(des_header.Security.CLAS.strip()))
            return False

        extracted_class = the_sidd.ProductCreation.Classification.classification
        if extracted_class != des_header.Security.CLAS.strip():
            logger.warning(
                'SIDD: DES.DESCLAS is `{}`,\n\tand SIDD.ProductCreation.Classification.classification '
                'is {}'.format(des_header.Security.CLAS.strip(), extracted_class))

        if des_header.Security.CLAS.strip() != nitf_details.nitf_header.Security.CLAS.strip():
            logger.warning(
                'SIDD: DES.DESCLAS is `{}`,\n\tand NITF.CLAS is `{}`'.format(
                    des_header.Security.CLAS.strip(), nitf_details.nitf_header.Security.CLAS.strip()))
        return True

    # check sidd xml structure for validity
    valid_sidd, xml_urn, the_sidd = _evaluate_xml_string_validity(xml_string)
    # check that the sidd information and header information appropriately match
    valid_header = check_des_header_fields()
    # check that the classification seems to make sense
    valid_class = compare_sidd_class()
    return valid_sidd & valid_header & valid_class, the_sidd


def check_sidd_file(nitf_details):
    """
    Check the validity of the given NITF file as a SIDD file.

    Parameters
    ----------
    nitf_details : str|NITFDetails

    Returns
    -------
    bool
    """

    def find_des():
        for i in range(nitf_details.des_subheader_offsets.size):
            subhead_bytes = nitf_details.get_des_subheader_bytes(i)
            if nitf_details.nitf_version == '02.00':
                des_header = DataExtensionHeader0.from_bytes(subhead_bytes, start=0)
            elif nitf_details.nitf_version == '02.10':
                des_header = DataExtensionHeader.from_bytes(subhead_bytes, start=0)
            else:
                raise ValueError('Got unhandled NITF version {}'.format(nitf_details.nitf_version))

            if subhead_bytes.startswith(b'DEXML_DATA_CONTENT'):
                des_bytes = nitf_details.get_des_bytes(i).decode('utf-8').strip().encode()
                # noinspection PyBroadException
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SIDD' in root_node.tag:  # namespace makes this ugly
                        sidd_des.append((i, des_bytes, root_node, xml_ns, des_header))
                    elif 'SICD' in root_node.tag:
                        sicd_des.append((i, des_bytes, root_node, xml_ns, des_header))
                except Exception:
                    continue
            elif subhead_bytes.startswith(b'DESIDD_XML'):
                # This is an old format SIDD
                des_bytes = nitf_details.get_des_bytes(i).decode('utf-8').strip().encode()
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SIDD' in root_node.tag:  # namespace makes this ugly
                        sidd_des.append((i, des_bytes, root_node, xml_ns, des_header))
                except Exception as e:
                    logger.exception('SIDD: Old-style SIDD DES header at index {}, but failed parsing'.format(i))
                    continue
            elif subhead_bytes.startswith(b'DESICD_XML'):
                # This is an old format SICD
                des_bytes = nitf_details.get_des_bytes(i).decode('utf-8').strip().encode()
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SICD' in root_node.tag:  # namespace makes this ugly
                        sicd_des.append((i, des_bytes, root_node, xml_ns, des_header))
                except Exception as e:
                    logger.exception('SIDD: Old-style SICD DES header at index {}, but failed parsing'.format(i))
                    continue


    def check_image_data():
        valid_images = True
        # verify that all images have the correct pixel type
        for i, img_header in enumerate(nitf_details.img_headers):
            if img_header.ICAT.strip() != 'SAR':
                continue

            iid1 = img_header.IID1.strip()
            if re.match(r'^SIDD\d\d\d\d\d\d', iid1) is None:
                valid_images = False
                logger.error(
                    'SIDD: image segment at index {} of {} has IID1 = `{}`,\n\t'
                    'expected to be of the form `SIDDXXXYYY`'.format(i, len(nitf_details.img_headers), iid1))
                continue

            sidd_index = int(iid1[4:7])
            if not (0 < sidd_index <= len(sidd_des)):
                valid_images = False
                logger.error(
                    'SIDD: image segment at index {} of {} has IID1 = `{}`,\n\t'
                    'it is unclear with which of the {} SIDDs '
                    'this is associated'.format(i, len(nitf_details.img_headers), iid1, len(sidd_des)))
                continue

            has_image[sidd_index - 1] = True

            type_information = sidd_nitf_details[sidd_index - 1]
            pixel_type = type_information['pixel_type']
            if pixel_type is None:
                continue  # we already noted the failure here

            exp_nbpp, exp_pvtype = type_information['nbpp'], type_information['pvtype']
            if img_header.PVTYPE.strip() != exp_pvtype:
                valid_images = False
                logger.error(
                    'SIDD: image segment at index {} of {} has PVTYPE = `{}`,\n\t'
                    'expected to be `{}` based on pixel type {}'.format(
                        i, len(nitf_details.img_headers), img_header.PVTYPE.strip(), exp_pvtype, pixel_type))
            if img_header.NBPP != exp_nbpp:
                valid_images = False
                logger.error(
                    'SIDD: image segment at index {} of {} has NBPP = `{}`,\n\t'
                    'expected to be `{}` based on pixel type {}'.format(
                        i, len(nitf_details.img_headers), img_header.NBPP, exp_nbpp, pixel_type))

        for sidd_index, entry in enumerate(has_image):
            if not entry:
                logger.error(
                    'SIDD: No image segments appear to be associated with the sidd at index {}'.format(sidd_index))
                valid_images = False

        return valid_images

    if isinstance(nitf_details, str):
        if not os.path.isfile(nitf_details):
            raise ValueError('Got string input, but it is not a valid path')
        nitf_details = NITFDetails(nitf_details)

    if not isinstance(nitf_details, NITFDetails):
        raise TypeError(
            'Input is expected to be a path to a NITF file, or a NITFDetails object instance')

    sidd_des = []
    sicd_des = []
    sidd_nitf_details = []
    has_image = []

    find_des()
    if len(sidd_des) < 1:
        logger.error('SIDD: No SIDD DES found, this is not a valid SIDD file.')
        return False

    valid_sidd_des = True
    for entry in sidd_des:
        this_sidd_valid, the_sidd = check_sidd_data_extension(nitf_details, entry[4], entry[1])
        valid_sidd_des &= this_sidd_valid
        has_image.append(False)
        if the_sidd.Display is None or the_sidd.Display.PixelType is None:
            valid_sidd_des = False
            logger.error('SIDD: SIDD.Display.PixelType is not populated, and can not be compared to NITF image details')
            sidd_nitf_details.append({'pixel_type': None})
        elif the_sidd.Display.PixelType in ['MONO8I', 'MONO8LU', 'RGB8LU']:
            sidd_nitf_details.append({'nbpp': 8, 'pvtype': 'INT', 'pixel_type': the_sidd.Display.PixelType})
        elif the_sidd.Display.PixelType == 'MONO16I':
            sidd_nitf_details.append({'nbpp': 16, 'pvtype': 'INT', 'pixel_type': the_sidd.Display.PixelType})
        elif the_sidd.Display.PixelType == 'RGB24I':
            # RGB24I is 3 channels of 8 bpp, see: https://github.com/ngageoint/sarpy/issues/544
            sidd_nitf_details.append({'nbpp': 8, 'pvtype': 'INT', 'pixel_type': the_sidd.Display.PixelType})
        else:
            raise ValueError('Got unhandled pixel type {}'.format(the_sidd.Display.PixelType))

    valid_sicd_des = True
    for entry in sicd_des:
        this_sicd_valid, _ = check_sicd_data_extension(nitf_details, entry[4], entry[1])
        valid_sicd_des &= this_sicd_valid
    valid_image = check_image_data()

    return valid_sidd_des & valid_sicd_des & valid_image


def check_file(file_name):
    """
    Check the validity for the given file SIDD (i.e. appropriately styled NITF)
    or xml file containing the SIDD structure alone.

    Parameters
    ----------
    file_name : str|NITFDetails

    Returns
    -------
    bool
    """

    if isinstance(file_name, str):
        if not os.path.isfile(file_name):
            raise ValueError('Got string input, but it is not a valid path')

        # check if this is just an xml file
        with open(file_name, 'rb') as fi:
            initial_bits = fi.read(30)
            if initial_bits.startswith(b'<?xml') or initial_bits.startswith(b'<SIDD'):
                fi.seek(0)
                sidd_xml = fi.read()
                return _evaluate_xml_string_validity(sidd_xml)[0]

    return check_sidd_file(file_name)


def main(args=None):
    parser = argparse.ArgumentParser('SIDD Consistency')
    parser.add_argument('file_name')
    parser.add_argument(
        '-l', '--level', default='WARNING',
        choices=['INFO', 'WARNING', 'ERROR'], help="Logging level")
    config = parser.parse_args(args)

    logging.basicConfig(level=config.level)
    logger.setLevel(config.level)
    validity = check_file(config.file_name)
    if validity:
        logger.info('\nSIDD: {} has been validated with no errors'.format(config.file_name))
    else:
        logger.error('\nSIDD: {} has apparent errors'.format(config.file_name))
    return int(validity)


if __name__ == '__main__':
    sys.exit(main())    # pragma: no cover



"""
The Channel definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List, Tuple, Optional

import numpy

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import POLARIZATION_TYPE, AreaType, AddedParametersType
from sarpy.io.xml.base import Serializable, SerializableArray, \
    Arrayable
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, StringListDescriptor, \
    IntegerDescriptor, FloatDescriptor, BooleanDescriptor, \
    SerializableDescriptor, SerializableListDescriptor, SerializableArrayDescriptor


class PolarizationRefType(Serializable, Arrayable):
    """
    Polarization reference type.
    """
    _fields = ('AmpH', 'AmpV', 'PhaseV')
    _required = _fields
    _numeric_format = {key: '0.17E' for key in _fields}
    AmpH = FloatDescriptor(
        'AmpH', _required, strict=DEFAULT_STRICT, bounds=(0.0, 1.0),
        docstring='E-field relative amplitude in H direction')  # type: float
    AmpV = FloatDescriptor(
        'AmpV', _required, strict=DEFAULT_STRICT, bounds=(0.0, 1.0),
        docstring='E-field relative amplitude in V direction')  # type: float
    PhaseV = FloatDescriptor(
        'PhaseV', _required, strict=DEFAULT_STRICT, bounds=(-0.5, 0.5),
        docstring='Relative phase of the V E-field '
                  'relative to the H E-field')  # type: float

    def __init__(
            self,
            AmpH: float = None,
            AmpV: float = None,
            PhaseV: float = None,
            **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.AmpH = AmpH
        self.AmpV = AmpV
        self.PhaseV = PhaseV
        super(PolarizationRefType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64) -> numpy.ndarray:
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [AmpH, AmpV, PhaseV]
        """

        return numpy.array([self.AmpH, self.AmpV, self.PhaseV], dtype=dtype)

    @classmethod
    def from_array(cls, array: numpy.ndarray):
        """
        Construct from an iterable.

        Parameters
        ----------
        array : numpy.ndarray|list|tuple

        Returns
        -------
        PolarizationRefType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) != 3:
                raise ValueError(
                    'Expected array to be of length 3,\n\t'
                    'and received `{}`'.format(array))
            return cls(AmpH=array[0], AmpV=array[1], PhaseV=array[2])
        raise ValueError(
            'Expected array to be numpy.ndarray, list, or tuple,\n\t'
            'got `{}`'.format(type(array)))


class PolarizationType(Serializable):
    """
    Polarization(s) of the signals that formed the signal array.
    """

    _fields = ('TxPol', 'RcvPol', 'TxPolRef', 'RcvPolRef')
    _required = ('TxPol', 'RcvPol')
    # descriptors
    TxPol = StringEnumDescriptor(
        'TxPol', POLARIZATION_TYPE, _required, strict=DEFAULT_STRICT,
        docstring='Transmitted signal polarization for the channel.')  # type: str
    RcvPol = StringEnumDescriptor(
        'RcvPol', POLARIZATION_TYPE, _required, strict=DEFAULT_STRICT,
        docstring='Receive polarization for the channel.')  # type: str
    TxPolRef = SerializableDescriptor(
        'TxPolRef', PolarizationRefType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[PolarizationRefType]
    RcvPolRef = SerializableDescriptor(
        'RcvPolRef', PolarizationRefType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[PolarizationRefType]

    def __init__(
            self,
            TxPol: str = None,
            RcvPol: str = None,
            TxPolRef: Union[None, PolarizationRefType, numpy.ndarray, tuple, list] = None,
            RcvPolRef: Union[None, PolarizationRefType, numpy.ndarray, tuple, list] = None,
            **kwargs):
        """

        Parameters
        ----------
        TxPol : str
        RcvPol : str
        TxPolRef : None|PolarizationRefType|numpy.ndarray|tuple|list
        RcvPolRef : None|PolarizationRefType|numpy.ndarray|tuple|list
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxPol = TxPol
        self.RcvPol = RcvPol
        self.TxPolRef = TxPolRef
        self.RcvPolRef = RcvPolRef
        super(PolarizationType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        for fld in ['TxPol', 'RcvPol']:
            val = getattr(self, fld)
            if val is not None and val in ['S', 'E']:
                required = max(required, (1, 1, 0))
        if self.TxPolRef is not None or self.RcvPolRef is not None:
            required = max(required, (1, 1, 0))
        return required


class LFMEclipseType(Serializable):
    """
    The LFM Eclipse definition.
    """

    _fields = ('FxEarlyLow', 'FxEarlyHigh', 'FxLateLow', 'FxLateHigh')
    _required = _fields
    _numeric_format = {fld: FLOAT_FORMAT for fld in _fields}
    # descriptors
    FxEarlyLow = FloatDescriptor(
        'FxEarlyLow', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring=r'FX domain minimum frequency value for an echo at '
                  r':math:`\Delta TOA = \Delta TOAE1 < \Delta TOA1`, in Hz.')  # type: float
    FxEarlyHigh = FloatDescriptor(
        'FxEarlyHigh', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='FX domain maximum frequency value for an echo at '
                  r':math:`\Delta TOA = \Delta TOAE1 < \Delta TOA1`, in Hz.')  # type: float
    FxLateLow = FloatDescriptor(
        'FxLateLow', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='FX domain minimum frequency value for an echo at '
                  r':math:`\Delta TOA = \Delta TOAE2 < \Delta TOA2`, in Hz.')  # type: float
    FxLateHigh = FloatDescriptor(
        'FxLateHigh', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='FX domain maximum frequency value for echo at '
                  r':math:`\Delta TOA = \Delta TOAE2 < \Delta TOA2`, in Hz.')  # type: float

    def __init__(self, FxEarlyLow=None, FxEarlyHigh=None, FxLateLow=None, FxLateHigh=None,
                 **kwargs):
        """

        Parameters
        ----------
        FxEarlyLow : float
        FxEarlyHigh : float
        FxLateLow : float
        FxLateHigh : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.FxEarlyLow = FxEarlyLow
        self.FxEarlyHigh = FxEarlyHigh
        self.FxLateLow = FxLateLow
        self.FxLateHigh = FxLateHigh
        super(LFMEclipseType, self).__init__(**kwargs)


class TOAExtendedType(Serializable):
    """
    The time-of-arrival (TOA) extended swath information.
    """

    _fields = ('TOAExtSaved', 'LFMEclipse')
    _required = ('TOAExtSaved', )
    _numeric_format = {'TOAExtSaved': FLOAT_FORMAT}
    # descriptors
    TOAExtSaved = FloatDescriptor(
        'TOAExtSaved', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='TOA extended swath saved that includes both full and partially '
                  'eclipsed echoes.')  # type: float
    LFMEclipse = SerializableDescriptor(
        'LFMEclipse', LFMEclipseType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that describe the FX domain signal content for partially '
                  'eclipsed echoes when the collection is performed with a Linear '
                  'FM waveform.')  # type: Union[None, LFMEclipseType]

    def __init__(self, TOAExtSaved=None, LFMEclipse=None, **kwargs):
        """

        Parameters
        ----------
        TOAExtSaved : float
        LFMEclipse : None|LFMEclipseType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TOAExtSaved = TOAExtSaved
        self.LFMEclipse = LFMEclipse
        super(TOAExtendedType, self).__init__(**kwargs)


class DwellTimesType(Serializable):
    """
    COD Time and Dwell Time polynomials over the image area.
    """

    _fields = ('CODId', 'DwellId', 'DTAId', 'UseDTA')
    _required = ('CODId', 'DwellId')
    # descriptors
    CODId = StringDescriptor(
        'CODId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of the Center of Dwell Time polynomial that maps '
                  'reference surface position to COD time.')  # type: str
    DwellId = StringDescriptor(
        'DwellId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of the Dwell Time polynomial that maps reference '
                  'surface position to dwell time.')  # type: str
    DTAId = StringDescriptor(
        'DTAId', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[str]
    UseDTA = BooleanDescriptor(
        'UseDTA', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[bool]

    def __init__(self, CODId=None, DwellId=None, DTAId=None, UseDTA=None, **kwargs):
        """

        Parameters
        ----------
        CODId : str
        DwellId : str
        DTAId : None|str
        UseDTA : None|bool
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CODId = CODId
        self.DwellId = DwellId
        self.DTAId = DTAId
        self.UseDTA = UseDTA
        super(DwellTimesType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        if self.DTAId is not None or self.UseDTA is not None:
            return (1, 1, 0)
        else:
            return (1, 0, 1)


class AntennaType(Serializable):
    """
    Antenna Phase Center and Antenna Pattern identifiers for
    the antenna(s) used to collect and form the signal array data.
    """

    _fields = ('TxAPCId', 'TxAPATId', 'RcvAPCId', 'RcvAPATId')
    _required = _fields
    # descriptors
    TxAPCId = StringDescriptor(
        'TxAPCId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of Transmit APC to be used to compute the transmit '
                  'antenna pattern as a function of time for the channel.')  # type: str
    TxAPATId = StringDescriptor(
        'TxAPATId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of Transmit Antenna pattern used to form the channel '
                  'signal array.')  # type: str
    RcvAPCId = StringDescriptor(
        'RcvAPCId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of Receive APC to be used to compute the receive antenna '
                  'pattern as a function of time for the channel.')  # type: str
    RcvAPATId = StringDescriptor(
        'RcvAPATId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of Receive Antenna pattern used to form the '
                  'channel.')  # type: str

    def __init__(self, TxAPCId=None, TxAPATId=None, RcvAPCId=None, RcvAPATId=None, **kwargs):
        """

        Parameters
        ----------
        TxAPCId : str
        TxAPATId : str
        RcvAPCId : str
        RcvAPATId : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxAPCId = TxAPCId
        self.TxAPATId = TxAPATId
        self.RcvAPCId = RcvAPCId
        self.RcvAPATId = RcvAPATId
        super(AntennaType, self).__init__(**kwargs)


class TxRcvType(Serializable):
    """
    Parameters to identify the Transmit and Receive parameter sets used to collect the signal array.
    """

    _fields = ('TxWFId', 'RcvId')
    _required = _fields
    _collections_tags = {
        'TxWFId': {'array': False, 'child_tag': 'TxWFId'},
        'RcvId': {'array': False, 'child_tag': 'RcvId'}}
    # descriptors
    TxWFId = StringListDescriptor(
        'TxWFId', _required, strict=DEFAULT_STRICT, minimum_length=1,
        docstring='Identifier of the Transmit Waveform parameter set(s) that '
                  'were used.')  # type: List[str]
    RcvId = StringListDescriptor(
        'RcvId', _required, strict=DEFAULT_STRICT, minimum_length=1,
        docstring='Identifier of the Receive Parameter set(s) that were '
                  'used.')  # type: List[str]

    def __init__(self, TxWFId=None, RcvId=None, **kwargs):
        """

        Parameters
        ----------
        TxWFId : List[str]
        RcvId : List[str]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxWFId = TxWFId
        self.RcvId = RcvId
        super(TxRcvType, self).__init__(**kwargs)


class TgtRefLevelType(Serializable):
    """
    Signal level for an ideal point scatterer located at the SRP for reference
    signal vector.
    """

    _fields = ('PTRef', )
    _required = _fields
    _numeric_format = {'PTRef': FLOAT_FORMAT}
    # descriptors
    PTRef = FloatDescriptor(
        'PTRef', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Target power level for a 1.0 square meter ideal point scatterer located '
                  'at the SRP. For FX Domain signal arrays, PTRef is the signal level at '
                  ':math:`fx = fx_C`. For TOA Domain, PTRef is the peak signal level at '
                  r':math:`\Delta TOA = 0`, and :math:`Power = |Signal|^2`.')  # type: float

    def __init__(self, PTRef=None, **kwargs):
        """

        Parameters
        ----------
        PTRef : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PTRef = PTRef
        super(TgtRefLevelType, self).__init__(**kwargs)


class FxPNPointType(Serializable):
    """
    Points that describe the noise profile.
    """

    _fields = ('Fx', 'PN')
    _required = _fields
    _numeric_format = {'FX': FLOAT_FORMAT, 'PN': FLOAT_FORMAT}
    # descriptors
    Fx = FloatDescriptor(
        'Fx', _required, strict=DEFAULT_STRICT,
        docstring='Frequency value of this noise profile point, in Hz.')  # type: float
    PN = FloatDescriptor(
        'PN', _required, strict=DEFAULT_STRICT,
        docstring='Power level of this noise profile point.')  # type: float

    def __init__(self, Fx=None, PN=None, **kwargs):
        """

        Parameters
        ----------
        Fx : float
        PN : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Fx = Fx
        self.PN = PN
        super(FxPNPointType, self).__init__(**kwargs)


class FxNoiseProfileType(SerializableArray):
    _set_size = False
    _set_index = False


class NoiseLevelType(Serializable):
    """
    Thermal noise level for the reference signal vector.
    """

    _fields = ('PNRef', 'BNRef', 'FxNoiseProfile')
    _required = ('PNRef', 'BNRef')
    _collections_tags = {
        'FxNoiseProfile': {'array': True, 'child_tag': 'Point'}}
    _numeric_format = {'PNRef': FLOAT_FORMAT, 'BNRef': FLOAT_FORMAT}
    # descriptors
    PNRef = FloatDescriptor(
        'PNRef', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Noise power level for thermal noise.')  # type: float
    BNRef = FloatDescriptor(
        'BNRef', _required, strict=DEFAULT_STRICT, bounds=(0, 1),
        docstring='Noise Equivalent BW for noise signal. Bandwidth BN is '
                  'expressed relative to the sample bandwidth.')  # type: float
    FxNoiseProfile = SerializableArrayDescriptor(
        'FxNoiseProfile', FxPNPointType, _collections_tags, _required, strict=DEFAULT_STRICT,
        minimum_length=2, array_extension=FxNoiseProfileType,
        docstring='FX Domain Noise Level Profile. Power level for thermal noise (PN) vs. FX '
                  'frequency values.')  # type: Union[None, FxNoiseProfileType, List[FxPNPointType]]

    def __init__(self, PNRef=None, BNRef=None, FxNoiseProfile=None, **kwargs):
        """

        Parameters
        ----------
        PNRef : float
        BNRef : float
        FxNoiseProfile : FxNoiseProfileType|List[FxPNPointType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PNRef = PNRef
        self.BNRef = BNRef
        self.FxNoiseProfile = FxNoiseProfile
        super(NoiseLevelType, self).__init__(**kwargs)


class ChannelParametersType(Serializable):
    _fields = (
        'Identifier', 'RefVectorIndex', 'FXFixed', 'TOAFixed', 'SRPFixed',
        'SignalNormal', 'Polarization', 'FxC', 'FxBW', 'FxBWNoise', 'TOASaved',
        'TOAExtended', 'DwellTimes', 'ImageArea', 'Antenna', 'TxRcv',
        'TgtRefLevel', 'NoiseLevel')
    _required = (
        'Identifier', 'RefVectorIndex', 'FXFixed', 'TOAFixed', 'SRPFixed',
        'Polarization', 'FxC', 'FxBW', 'TOASaved', 'DwellTimes')
    _numeric_format = {
        'FxC': FLOAT_FORMAT, 'FxBW': FLOAT_FORMAT, 'FxBWNoise': FLOAT_FORMAT, 'TOASaved': FLOAT_FORMAT}
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this CPHD data channel.')  # type: str
    RefVectorIndex = IntegerDescriptor(
        'RefVectorIndex', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Index of the reference vector for the channel.')  # type: int
    FXFixed = BooleanDescriptor(
        'FXFixed', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate when a constant FX band is saved for all signal '
                  'vectors of the channel.')  # type: bool
    TOAFixed = BooleanDescriptor(
        'TOAFixed', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate when a constant TOA swath is saved for all '
                  'signal vectors of the channel.')  # type: bool
    SRPFixed = BooleanDescriptor(
        'SRPFixed', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate when a constant SRP position is used all '
                  'signal vectors of the channel.')  # type: bool
    SignalNormal = BooleanDescriptor(
        'SignalNormal', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate when all signal array vectors are normal. '
                  'Included if and only if the SIGNAL PVP is also included.')  # type: bool
    Polarization = SerializableDescriptor(
        'Polarization', PolarizationType, _required, strict=DEFAULT_STRICT,
        docstring='Polarization(s) of the signals that formed the signal '
                  'array.')  # type: PolarizationType
    FxC = FloatDescriptor(
        'FxC', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='FX center frequency value for saved bandwidth for the channel. '
                  'Computed from all vectors of the signal array.')  # type: float
    FxBW = FloatDescriptor(
        'FxBW', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='FX band spanned for the saved bandwidth for the channel. '
                  'Computed from all vectors of the signal array.')  # type: float
    FxBWNoise = FloatDescriptor(
        'FxBWNoise', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='FX signal bandwidth saved that includes noise signal below or '
                  'above the retained echo signal bandwidth.')  # type: float
    TOASaved = FloatDescriptor(
        'TOASaved', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='TOA swath saved for the full resolution echoes for the channel.')  # type: float
    TOAExtended = SerializableDescriptor(
        'TOAExtended', TOAExtendedType, _required, strict=DEFAULT_STRICT,
        docstring='TOA extended swath information.')  # type: Union[None, TOAExtendedType]
    DwellTimes = SerializableDescriptor(
        'DwellTimes', DwellTimesType, _required, strict=DEFAULT_STRICT,
        docstring='COD Time and Dwell Time polynomials over the image area.')  # type: DwellTimesType
    ImageArea = SerializableDescriptor(
        'ImageArea', AreaType, _required, strict=DEFAULT_STRICT,
        docstring='Image Area for the CPHD channel defined by a rectangle aligned with '
                  '(IAX, IAY). May be reduced by the optional '
                  'polygon.')  # type: Union[None, AreaType]
    Antenna = SerializableDescriptor(
        'Antenna', AntennaType, _required, strict=DEFAULT_STRICT,
        docstring='Antenna Phase Center and Antenna Pattern identifiers for the antenna(s) '
                  'used to collect and form the signal array data.')  # type: Union[None, AntennaType]
    TxRcv = SerializableDescriptor(
        'TxRcv', TxRcvType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters to identify the Transmit and Receive parameter sets '
                  'used to collect the signal array.')  # type: Union[None, TxRcvType]
    TgtRefLevel = SerializableDescriptor(
        'TgtRefLevel', TgtRefLevelType, _required, strict=DEFAULT_STRICT,
        docstring='Signal level for an ideal point scatterer located at the SRP for '
                  'reference signal vector.')  # type: Union[None, TgtRefLevelType]
    NoiseLevel = SerializableDescriptor(
        'NoiseLevel', NoiseLevelType, _required, strict=DEFAULT_STRICT,
        docstring='Thermal noise level for the reference signal '
                  'vector.')  # type: Union[None, NoiseLevelType]

    def __init__(self, Identifier=None, RefVectorIndex=None, FXFixed=None, TOAFixed=None,
                 SRPFixed=None, SignalNormal=None, Polarization=None, FxC=None, FxBW=None,
                 FxBWNoise=None, TOASaved=None, TOAExtended=None, DwellTimes=None,
                 ImageArea=None, Antenna=None, TxRcv=None, TgtRefLevel=None,
                 NoiseLevel=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        RefVectorIndex : int
        FXFixed : bool
        TOAFixed : bool
        SRPFixed : bool
        SignalNormal : None|bool
        Polarization : PolarizationType
        FxC : float
        FxBW : float
        FxBWNoise : None|float
        TOASaved : float
        TOAExtended : None|TOAExtendedType
        DwellTimes : DwellTimesType
        ImageArea : None|AreaType
        Antenna : None|AntennaType
        TxRcv : None|TxRcvType
        TgtRefLevel : None|TgtRefLevelType
        NoiseLevel : None|NoiseLevelType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.RefVectorIndex = RefVectorIndex
        self.FXFixed = FXFixed
        self.TOAFixed = TOAFixed
        self.SRPFixed = SRPFixed
        self.SignalNormal = SignalNormal
        self.Polarization = Polarization
        self.FxC = FxC
        self.FxBW = FxBW
        self.FxBWNoise = FxBWNoise
        self.TOASaved = TOASaved
        self.TOAExtended = TOAExtended
        self.DwellTimes = DwellTimes
        self.ImageArea = ImageArea
        self.Antenna = Antenna
        self.TxRcv = TxRcv
        self.TgtRefLevel = TgtRefLevel
        self.NoiseLevel = NoiseLevel
        super(ChannelParametersType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.Polarization is not None:
            required = max(required, self.Polarization.version_required())
        if self.DwellTimes is not None:
            required = max(required, self.DwellTimes.version_required())
        return required


class ChannelType(Serializable):
    """
    The channel definition.
    """

    _fields = (
        'RefChId', 'FXFixedCPHD', 'TOAFixedCPHD', 'SRPFixedCPHD',
        'Parameters', 'AddedParameters')
    _required = (
        'RefChId', 'FXFixedCPHD', 'TOAFixedCPHD', 'SRPFixedCPHD', 'Parameters')
    _collections_tags = {
        'Parameters': {'array': False, 'child_tag': 'Parameters'}}
    # descriptors
    RefChId = StringDescriptor(
        'RefChId', _required, strict=DEFAULT_STRICT,
        docstring='Channel ID for the Reference Channel in the '
                  'product.')  # type: str
    FXFixedCPHD = BooleanDescriptor(
        'FXFixedCPHD', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate when a constant FX band is saved for all '
                  'signal vectors of all channels.')  # type: bool
    TOAFixedCPHD = BooleanDescriptor(
        'TOAFixedCPHD', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate when a constant TOA swath is saved for all '
                  'signal vectors of all channels.')  # type: bool
    SRPFixedCPHD = BooleanDescriptor(
        'SRPFixedCPHD', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate when a constant SRP position is used all '
                  'signal vectors of all channels.')  # type: bool
    Parameters = SerializableListDescriptor(
        'Parameters', ChannelParametersType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameter Set that describes a CPHD data '
                  'channel.')  # type: List[ChannelParametersType]
    AddedParameters = SerializableDescriptor(
        'AddedParameters', AddedParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Block for including additional parameters that describe '
                  'the channels and/or signal arrays.')  # type: Union[None, AddedParametersType]

    def __init__(
            self,
            RefChId: str = None,
            FXFixedCPHD: bool = None,
            TOAFixedCPHD: bool = None,
            SRPFixedCPHD: bool = None,
            Parameters: List[ChannelParametersType] = None,
            AddedParameters: Optional[AddedParametersType] = None,
            **kwargs):
        """

        Parameters
        ----------
        RefChId : str
        FXFixedCPHD : bool
        TOAFixedCPHD : bool
        SRPFixedCPHD : bool
        Parameters : List[ChannelParametersType]
        AddedParameters : None|AddedParametersType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RefChId = RefChId
        self.FXFixedCPHD = FXFixedCPHD
        self.TOAFixedCPHD = TOAFixedCPHD
        self.SRPFixedCPHD = SRPFixedCPHD
        self.Parameters = Parameters
        self.AddedParameters = AddedParameters
        super(ChannelType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.Parameters is not None:
            for entry in self.Parameters:
                required = max(required, entry.version_required())
        return required



"""
The Channel definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Michael Stewart, Valkyrie")

from typing import Union, List

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, \
    IntegerDescriptor, SerializableListDescriptor, ParametersDescriptor, \
    BooleanDescriptor, FloatDescriptor, SerializableDescriptor

from sarpy.io.phase_history.cphd1_elements.blocks import POLARIZATION_TYPE, AreaType
from sarpy.io.phase_history.cphd1_elements.Channel import DwellTimesType

from .base import DEFAULT_STRICT, FLOAT_FORMAT


class RcvAntennaType(Serializable):
    """
    The receive antenna information.
    """

    _fields = ('RcvAPCId', 'RcvAPATId')
    _required = _fields
    # descriptors
    RcvAPCId = StringDescriptor(
        'RcvAPCId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier for the Receive APC to be used to compute the receive antenna'
                  ' pattern as a function of time for the channel (APC_ID).')  # type: str
    RcvAPATId = StringDescriptor(
        'RcvAPATId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier for the Receive Antenna pattern to collect the signal data'
                  ' (APAT_ID).')  # type: str

    def __init__(self, RcvAPCId=None, RcvAPATId=None, **kwargs):
        """

        Parameters
        ----------
        RcvAPCId : str
        RcvAPATId : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RcvAPCId = RcvAPCId
        self.RcvAPATId = RcvAPATId
        super(RcvAntennaType, self).__init__(**kwargs)


class SignalRefLevelType(Serializable):
    """
    The signal power level information.
    """

    _fields = ('PSCRSD', 'PRcvDensity')
    _required = _fields
    _numeric_format = {fld: FLOAT_FORMAT for fld in _fields}
    # descriptors
    PSCRSD = FloatDescriptor(
        'PSCRSD', _required, strict=DEFAULT_STRICT,
        docstring='Power level in the fast time signal vector for a CW tone at f = f_0_REF'
                  ' and for f_IC(v,t) = f_0_REF.')  # type: float
    PRcvDensity = FloatDescriptor(
        'PRcvDensity', _required, strict=DEFAULT_STRICT,
        docstring='Receive power density per unit area for a CW tone at f = f_0_REF that results in'
                  ' signal vector power PS_CRSD. Signal received from a far field source located'
                  ' along the receive antenna mainlobe boresight at t = trs(v_CH_REF).')  # type:float

    def __init__(self, PSCRSD=None, PRcvDensity=None, **kwargs):
        """

        Parameters
        ----------
        PSCRSD : float
        PRcvDensity : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PSCRSD = PSCRSD
        self.PRcvDensity = PRcvDensity
        super(SignalRefLevelType, self).__init__(**kwargs)


class NoiseLevelType(Serializable):
    """
    The thermal noise level information.
    """

    _fields = ('PNCRSD', 'BNCRSD')
    _required = _fields
    _numeric_format = {fld: FLOAT_FORMAT for fld in _fields}
    # descriptors
    PNCRSD = FloatDescriptor(
        'PNCRSD', _required, strict=DEFAULT_STRICT,
        docstring='Noise power level in fast time signal vector for f_IC(v,t) = f_0(v_CH_REF).')  # type: float
    BNCRSD = FloatDescriptor(
        'BNCRSD', _required, strict=DEFAULT_STRICT,
        docstring='Noise Equivalent BW for the noise signal. Bandwidth BN_CRSD is expressed relative'
                  ' to the fast time sample rate for the channel (fs).')  # type:float

    def __init__(self, PNCRSD=None, BNCRSD=None, **kwargs):
        """

        Parameters
        ----------
        PNCRSD : float
        BNCRSD : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PNCRSD = PNCRSD
        self.BNCRSD = BNCRSD
        super(NoiseLevelType, self).__init__(**kwargs)


class TxAntennaType(Serializable):
    """
    The receive antenna information.
    """

    _fields = ('TxAPCId', 'TxAPATId')
    _required = _fields
    # descriptors
    TxAPCId = StringDescriptor(
        'TxAPCId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of Transmit APC to be used to compute the transmit antenna'
                  ' pattern as a function of time for the channel (APC_ID).')  # type: str
    TxAPATId = StringDescriptor(
        'TxAPATId', _required, strict=DEFAULT_STRICT,
        docstring='Identifier of the Transmit Antenna pattern used to form the channel'
                  ' signal array (APAT_ID).')  # type: str

    def __init__(self, TxAPCId=None, TxAPATId=None, **kwargs):
        """

        Parameters
        ----------
        TxAPCId : str
        TxAPATId : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxAPCId = TxAPCId
        self.TxAPATId = TxAPATId
        super(TxAntennaType, self).__init__(**kwargs)


class SARImagingType(Serializable):
    """
    The SAR Imaging parameters
    """

    _fields = ('TxLFMFixed', 'TxPol', 'DwellTimes', 'TxAntenna', 'ImageArea')
    _required = ('TxPol', 'DwellTimes')
    # descriptors
    TxLFMFixed = BooleanDescriptor(
        'TxLFMFixed', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate the same transmit LFM waveform is used for all pulses.'
                  ' vectors of the channel.')  # type: Union[None, bool]
    TxPol = StringEnumDescriptor(
        'TxPol', POLARIZATION_TYPE, _required, strict=DEFAULT_STRICT,
        docstring='Transmitted signal polarization for the channel.')  # type: str
    DwellTimes = SerializableDescriptor(
        'DwellTimes', DwellTimesType, _required, strict=DEFAULT_STRICT,
        docstring='COD Time and Dwell Time polynomials over the image area.')  # type: DwellTimesType
    TxAntenna = SerializableDescriptor(
        'TxAntenna', TxAntennaType, _required, strict=DEFAULT_STRICT,
        docstring='Antenna Phase Center and Antenna Pattern identifiers for the transmit antenna'
                  ' used to illuminate the imaged area.')  # type: Union[None, TxAntennaType]
    ImageArea = SerializableDescriptor(
        'ImageArea', AreaType, _required, strict=DEFAULT_STRICT,
        docstring='SAR Image Area for the channel defined by a rectangle aligned with (IAX, IAY).'
                  ' May be reduced by the optional polygon.')  # type: Union[None, AreaType]

    def __init__(self, TxLFMFixed=None, TxPol=None, DwellTimes=None, TxAntenna=None,
                 ImageArea=None, **kwargs):
        """

        Parameters
        ----------
        TxLFMFixed : None|bool
        TxPol : PolarizationType
        DwellTimes : DwellTimesType
        TxAntenna : None|TxAntennaType
        ImageArea : None|AreaType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxLFMFixed = TxLFMFixed
        self.TxPol = TxPol
        self.DwellTimes = DwellTimes
        self.TxAntenna = TxAntenna
        self.ImageArea = ImageArea
        super(SARImagingType, self).__init__(**kwargs)


class ChannelParametersType(Serializable):
    """
    The CRSD data channel parameters
    """
    _fields = (
        'Identifier', 'RefVectorIndex', 'RefFreqFixed', 'FrcvFixed', 'DemodFixed',
        'F0Ref', 'Fs', 'BWInst', 'RcvPol', 'SignalNormal', 'RcvAntenna',
        'SignalRefLevel', 'NoiseLevel', 'AddedParameters', 'SARImaging')
    _required = (
        'Identifier', 'RefVectorIndex', 'RefFreqFixed', 'FrcvFixed', 'DemodFixed',
        'F0Ref', 'Fs', 'BWInst', 'RcvPol')
    _numeric_format = {
        'F0Ref': FLOAT_FORMAT, 'Fs': FLOAT_FORMAT, 'BWInst': FLOAT_FORMAT}
    _collections_tags = {
        'AddedParameters': {'array': False, 'child_tag': 'AddedParameters'}}
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies this CRSD data channel.')  # type: str
    RefVectorIndex = IntegerDescriptor(
        'RefVectorIndex', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Index of the reference vector for the channel.')  # type: int
    RefFreqFixed = BooleanDescriptor(
        'RefFreqFixed', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate a constant reference frequency is used for'
                  ' the channel.')  # type: bool
    FrcvFixed = BooleanDescriptor(
        'FrcvFixed', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate a constant receive band is saved for the'
                  ' channel.')  # type: bool
    DemodFixed = BooleanDescriptor(
        'DemodFixed', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate a constant demodulation is used for the'
                  ' channel.')  # type: bool
    F0Ref = FloatDescriptor(
        'F0Ref', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Reference frequency for the reference signal vector.')  # type: float
    Fs = FloatDescriptor(
        'Fs', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Fast time sample rate for the signal array.')  # type: float
    BWInst = FloatDescriptor(
        'BWInst', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Nominal instantaneous bandwidth for the channel.')  # type: float
    RcvPol = StringEnumDescriptor(
        'RcvPol', POLARIZATION_TYPE, _required, strict=DEFAULT_STRICT,
        docstring='Receive polarization for the signal data processed to form the signal array.'
                  ' Parameter describes the E-Field orientation of the signal.')  # type: str
    SignalNormal = BooleanDescriptor(
        'SignalNormal', _required, strict=DEFAULT_STRICT,
        docstring='Flag to indicate when all signal array vectors are normal.'
                  ' Included if and only if the SIGNAL PVP is also included.')  # type: Union[None, bool]
    RcvAntenna = SerializableDescriptor(
        'RcvAntenna', RcvAntennaType, _required, strict=DEFAULT_STRICT,
        docstring='Antenna Phase Center and Antenna Pattern identifiers for the receive antenna'
                  ' used to collect and form the signal array data.')  # type: Union[None, RcvAntennaType]
    SignalRefLevel = SerializableDescriptor(
        'SignalRefLevel', SignalRefLevelType, _required, strict=DEFAULT_STRICT,
        docstring='Signal power levels for a received CW signal with f = f_0_REF and polarization'
                  ' matched to RcvPol of the channel.')  # type: Union[None, SignalRefLevelType]
    NoiseLevel = SerializableDescriptor(
        'NoiseLevel', NoiseLevelType, _required, strict=DEFAULT_STRICT,
        docstring='Thermal noise level in the CRSD signal vector for f_IC(v,t) ='
                  ' f_0(v_CH_REF).')  # type: Union[None, NoiseLevelType]
    AddedParameters = ParametersDescriptor(
        'AddedParameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Additional free form parameters.')  # type: Union[None, ParametersCollection]
    SARImaging = SerializableDescriptor(
        'SARImaging', SARImagingType, _required, strict=DEFAULT_STRICT,
        docstring='Structure included for all SAR imaging collections.')  # type: Union[None, SARImagingType]

    def __init__(self, Identifier=None, RefVectorIndex=None, RefFreqFixed=None,
                 FrcvFixed=None, DemodFixed=None, F0Ref=None, Fs=None, BWInst=None,
                 RcvPol=None, SignalNormal=None, RcvAntenna=None, SignalRefLevel=None,
                 NoiseLevel=None, AddedParameters=None, SARImaging=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        RefVectorIndex : int
        RefFreqFixed : bool
        FrcvFixed : bool
        DemodFixed : bool
        F0Ref : float
        Fs : float
        BWInst : float
        RcvPol : str
        SignalNormal : None|bool
        RcvAntenna : None|RcvAntennaType
        SignalRefLevel : None|SignalRefLevelType
        NoiseLevel : None|NoiseLevelType
        AddedParameters : None|ParametersCollection
        SARImaging : None|SARImagingType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.RefVectorIndex = RefVectorIndex
        self.RefFreqFixed = RefFreqFixed
        self.FrcvFixed = FrcvFixed
        self.DemodFixed = DemodFixed
        self.F0Ref = F0Ref
        self.Fs = Fs
        self.BWInst = BWInst
        self.RcvPol = RcvPol
        self.SignalNormal = SignalNormal
        self.RcvAntenna = RcvAntenna
        self.SignalRefLevel = SignalRefLevel
        self.NoiseLevel = NoiseLevel
        self.AddedParameters = AddedParameters
        self.SARImaging = SARImaging
        super(ChannelParametersType, self).__init__(**kwargs)


class ChannelType(Serializable):
    """
    The channel definition.
    """

    _fields = ('RefChId', 'Parameters')
    _required = _fields
    _collections_tags = {'Parameters': {'array': False, 'child_tag': 'Parameters'}}
    # descriptors
    RefChId = StringDescriptor(
        'RefChId', _required, strict=DEFAULT_STRICT,
        docstring='Channel ID for the Reference Channel in the '
                  'product.')  # type: str
    Parameters = SerializableListDescriptor(
        'Parameters', ChannelParametersType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameter Set that describes a CRSD data '
                  'channel.')  # type: List[ChannelParametersType]

    def __init__(self, RefChId=None, Parameters=None, **kwargs):
        """

        Parameters
        ----------
        RefChId : str
        Parameters : List[ChannelParametersType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RefChId = RefChId
        self.Parameters = Parameters
        super(ChannelType, self).__init__(**kwargs)
